/* 

(C) Copyright 2017 by Armin Biere
(C) Copyright 2005,2006 by Armin Biere
(C) Copyright 2006 by Toni Jussila

Institute for Formal Models and Verification
Johannes Kepler University
Linz, Austria

All rights reserved. Redistribution and use in source and binary forms, with
or without modification, are permitted provided that the following
conditions are met:

  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.

  3. All advertising materials mentioning features or use of this software
     must display the following acknowledgement:

    This product includes software developed by Armin Biere and Toni Jussila
    Institute for Formal Models and Verification, Johannes Kepler
    University, Linz, Austria.

  4. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.
   
THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "config.h"

/*------------------------------------------------------------------------*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/unistd.h>
#include <ctype.h>
#include <limits.h>

/*------------------------------------------------------------------------*/

#define SATISFIABLE 10
#define UNSATISFIABLE 20

/*------------------------------------------------------------------------*/

#define TRUE_AND INT_MAX
#define TRUE_OR INT_MAX
#define TRUE_EX INT_MAX

/*------------------------------------------------------------------------*/

#define CLAUSE 0
#define AND 1
#define EX 2

/*------------------------------------------------------------------------*/

/*

This code was developed on machines with less than 16GB main memory,
even though in principle there is no low-level bit-fiddling (except
maybe for computing hash functions), which could not easily be ported
to 64 bit machines.

Now that we have more memory in almost all of our machines this code
breaks pretty fast say for PHP20 within 200 seconds filling all the
32 bit index spaces we use.

Thus in order to at least give better error messages (instead of
segmentation fault) we introduced this MAXSIZE macro which the maximum
sizes of arrays indexed by (signed) 32-bit integers.

*/

#define MAXSIZE (int)(1<<30)

/*------------------------------------------------------------------------*/

typedef struct Ele Ele;
typedef struct And And;
typedef struct Ex Ex;
typedef struct Clause Clause;
typedef struct Line Line;
typedef struct Node Node;
typedef struct Var Var;

/*------------------------------------------------------------------------*/

typedef long long int64;

/*------------------------------------------------------------------------*/

typedef void (*InitOrder) (void);

/*------------------------------------------------------------------------*/

struct Ele
{
  int type;			/* clause, and or or */
  int idx;			/* index in corresponding table */
};

/*------------------------------------------------------------------------*/

struct Clause
{
  int bdd;
  int lits;
};

/*------------------------------------------------------------------------*/

struct Node
{
  int var;			/* variable of this Node */
  int child[2];			/* else (0), and then (1) child index */
};

/*------------------------------------------------------------------------*/

struct Line
{
  int f;			/* f < g */
  int g;			/* !is_trivial_step (f, g) */
  int h;			/* conjunction/disjunction of a AND b */
};

/*------------------------------------------------------------------------*/

struct Var
{
  int assignment;
  int occurrences;
  int bucket;
  int count;
  int bcnt;
};

/*------------------------------------------------------------------------*/

struct And
{
  int a;			/* eles index */
  int b;			/* eles index */
  int h;			/* bdds index */
};

/*------------------------------------------------------------------------*/

struct Ex
{
  int a;			/* eles index */
  int h;			/* bdds index */
};

/*------------------------------------------------------------------------*/

static FILE *trace_file;
static int traced_chains;
static int traced_variables;
static int extended_trace_format;

/*------------------------------------------------------------------------*/

static FILE *order_file;

/*------------------------------------------------------------------------*/

static int true_variable;
static int first_node_variable;

/*------------------------------------------------------------------------*/
/* The first part of the trace lists the original clauses as chains.
 */
static int first_clause_chain;

/* To simplify the trace generation we use the variable 'true_variable' for
 * true bdds and its negation for false bdds.  This variable is forced to
 * be true by adding a unit clause with (external) index 'true_clause'.
 */
static int true_chain;

/* Then we have the clauses that define the consistency constraints for
 * individual bdd nodes.  Since each bdd node is simply an 'if-then-else'
 * gate, we need four clauses for each BDD node.
 */
static int first_node_chain;

/* This part lists the unit clauses that are derived from individual
 * original clauses by defining a BDD for them.   Since the unit can be
 * derived by regular input resolution from the original clause and the
 * consistency constraints of the BDD nodes, only one new clause (actually
 * called 'chain' in 'tracecheck' terminology) is needed for each original
 * clause.
 */
static int first_root_unit_chain;

/* We add two clauses for each line in the cache.  The goal is to prove
 * 'a & b -> res' using Shannon's expansion rule.  However, as it
 * turns out this can not be done by regular input resolution.  Essentially
 * we need a case split, and prove 'a & b -> res | x', where 'x' is
 * the top variable with respect to 'a' and 'b'.  After obtaining
 * this clause a regular input resolution proof is possible.
 */
static int first_and_line_chain;

static int first_or_line_chain;
static int first_ex_chain;
static int first_ex_unit_chain;

/* This is the chain that is produced for the last BDD.  It has unit chains
 * representing bdds of original clauses and the clauses for the lines used
 * to build the last BDD as antecedents.
 */
static int root_chain;

/*------------------------------------------------------------------------*/

static FILE *output;
static int close_output;
static int num_buffer;
static char buffer[80];

/*------------------------------------------------------------------------*/

static FILE *input;
static const char *input_name;
static int close_input;
static int lineno;

/*------------------------------------------------------------------------*/

static size_t max_bytes;
static size_t current_bytes;

/*------------------------------------------------------------------------*/

static int max_variable;
static Var *variables;		/* [1 .. max_variable] */
static int *variable_order;	/* [1 .. max_variable] */
static int *variable_elim;	/* [1 .. max_variable] */
static int64 *variable_widths;
static int64 variable_cut;

/*------------------------------------------------------------------------*/

static int size_literals;
static int num_literals;
static int *literals;		/* [0 .. num_literals[ */

/*------------------------------------------------------------------------*/

static int size_occurrences;
static int *occurrences;	/* [0 .. size_occurrences[ */

/*------------------------------------------------------------------------*/

static int size_buckets;
static int *buckets;		/* [0 .. size_buckets[ */

/*------------------------------------------------------------------------*/

static int size_clauses;
static int num_clauses;
static Clause *clauses;		/* [1 .. num_clauses] */
static Clause *empty_clause;

/*------------------------------------------------------------------------*/
/* 0 = lin comp
 * 1 = tree comp 
 * 2 = ve
 * 3 = vetree
 */
static int strategy;

/*------------------------------------------------------------------------*/

static int *schedule;		/* [0 .. num_clauses [ */
static int orig_schedule = 1;
static int64 *clause_pos;	/* [1 .. num_clauses ] */

/*------------------------------------------------------------------------*/

static int num_nodes;
static int size_nodes;
static Node *nodes;		/* [2 .. num_nodes[ */

/*------------------------------------------------------------------------*/

static int num_table;
static int size_table;
static int *table;		/* [0 .. size_table[ */

/*------------------------------------------------------------------------*/

static int num_lines;
static int num_redundant_lines;
static int size_and_lines;
static Line *and_lines;		/* [1 .. num_lines] */

/*------------------------------------------------------------------------*/

static int num_or_lines;
static int num_redundant_or_lines;
static int size_or_lines;
static Line *or_lines;		/* [1 .. num_or_lines] */

/*------------------------------------------------------------------------*/

static int num_and_cache;
static int size_and_cache;
static int *and_cache;		/* [0 .. size_and_cache[ */

/*------------------------------------------------------------------------*/

static int num_or_cache;
static int size_or_cache;
static int *or_cache;		/* [0 .. size_or_cache[ */

/*------------------------------------------------------------------------*/

static int num_eles;
static int size_eles;
static Ele *eles;		/* [1 .. num_eles[ */

/*------------------------------------------------------------------------*/

static int num_ands;
static int num_trivial_ands;
static int num_redundant_ands;
static int size_ands;
static And *ands;		/* [0 .. num_ands[ */
static int root;

/*------------------------------------------------------------------------*/

static int num_trivial_exs;
static int num_redundant_exs;
static int num_exs;
static int size_exs;
static Ex *exs;			/* [0 .. max_var] */

/*------------------------------------------------------------------------*/

static int accu;

/*------------------------------------------------------------------------*/

static int elim_tresh;
static int trial_tresh = 2048;

/*------------------------------------------------------------------------*/

static int no_core;
static int num_core_ands;
static int num_core_exs;
static int num_core_lines;
static int num_core_or_lines;
static int num_core_nodes;
static int num_core_clauses;
static int num_core_variables;
static int num_trivial_core_ands;
static int num_trivial_core_exs;
static int num_redundant_core_ands;

/*------------------------------------------------------------------------*/

static double num_steps;

static double cache_finds;
static double cache_collisions;

static double cache_lookups;
static double cache_hits;

/*------------------------------------------------------------------------*/

static double table_collisions;
static double table_finds;

/*------------------------------------------------------------------------*/

static int size_marked_eles;	/* in 'unsigned' not bits */
static unsigned *marked_eles;	/* bit set */

/*------------------------------------------------------------------------*/

static int size_marked_ands;	/* in 'unsigned' not bits */
static unsigned *marked_ands;	/* bit set */

/*------------------------------------------------------------------------*/

static int size_marked_exs;	/* in 'unsigned' not bits */
static unsigned *marked_exs;	/* bit set */

/*------------------------------------------------------------------------*/

static int size_marked_clauses;	/* in 'unsigned' not bits */
static unsigned *marked_clauses;	/* bit set */

/*------------------------------------------------------------------------*/

static int size_marked_nodes;	/* in 'unsigned' not bits */
static unsigned *marked_nodes;	/* bit set */

/*------------------------------------------------------------------------*/

static int size_marked_and_lines;	/* in 'unsigned' not bits */
static unsigned *marked_and_lines;	/* bit set */

/*------------------------------------------------------------------------*/

static int size_marked_or_lines;	/* in 'unsigned' not bits */
static unsigned *marked_or_lines;	/* bit set */

/*------------------------------------------------------------------------*/

static int size_marked_variables;
static unsigned *marked_variables;

/*------------------------------------------------------------------------*/

static int size_stack;
static int num_stack;
static int *stack;		/* [0 ... count_stack[ */

/*------------------------------------------------------------------------*/

static double start_time;
static double sat_time;
static double core_time;
static double trace_time;
static double last_report_num_steps;
static int reports;
#ifdef EBDDRES_DUMP
static int dump_all;
#endif

/*------------------------------------------------------------------------*/

static double
current_time (void)
{
  struct rusage u;
  double res = 0;

  if (!getrusage (RUSAGE_SELF, &u))
    {
      res += u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
      res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
    }

  return res;
}

/*------------------------------------------------------------------------*/

static int
die (const char *fmt, ...)
{
  va_list ap;

  fputs ("*** ebddres: ", output);
  va_start (ap, fmt);
  vfprintf (output, fmt, ap);
  va_end (ap);
  fputc ('\n', output);
  fflush (output);
  abort ();
}

/*------------------------------------------------------------------------*/

#define NEW(P,BYTES) \
do { \
  if (!(BYTES)) { (P) = 0; break; } \
  (P) = (void*) malloc ((BYTES)); \
  if (!(P)) \
    die ("out-of-memory line %ld in '%s' allocating %ld bytes", \
      __LINE__, __FUNCTION__, (long)(BYTES)); \
  assert ((P)); \
  current_bytes += (BYTES); \
  if (current_bytes > max_bytes) \
    max_bytes = current_bytes; \
  memset ((P), 0, (BYTES)); \
} while (0)

/*------------------------------------------------------------------------*/

static void
delete (void *ptr, size_t bytes)
{
  assert (current_bytes >= bytes);
  current_bytes -= bytes;
  free (ptr);
}

/*------------------------------------------------------------------------*/

#define NEWN(p,n) do { NEW ((p), sizeof(*(p)) * (n)); } while (0)
#define DELETEN(p,n) do { delete ((p), sizeof(*(p)) * (n)); } while (0)

/*------------------------------------------------------------------------*/

#define REALLOCN(p,o,n) \
  do { \
    void * q; /* = new (sizeof(*(p)) * (n)); */ \
    NEW ((q), sizeof(*(p)) * (n)); \
    assert ((n) >= (o)); \
    assert ((q)); \
    memcpy (q, (p), sizeof (*(p)) * ((n) - (o))); \
    DELETEN(p,o); \
    (p) = q; \
  } while (0)

/*------------------------------------------------------------------------*/

static void
push (int elem)
{
  if (num_stack == size_stack)
    {
      if (size_stack == MAXSIZE) die ("stack size too large");
      size_stack *= 2;
      REALLOCN (stack, num_stack, size_stack);
    }

  stack[num_stack++] = elem;
}

/*------------------------------------------------------------------------*/

static int
pop (void)
{
  assert (num_stack > 0);
  return stack[--num_stack];
}

/*------------------------------------------------------------------------*/

static int
next (void)
{
  int res = getc (input);
  if (res == '\n')
    lineno++;
  return res;
}

/*------------------------------------------------------------------------*/

static const char *
parse (void)
{
  int v, c, ch, sign, lit, lits;
  Clause *clause;
  Ele *ele;

COMMENTS:
  ch = next ();

  if (ch == 'c')
    {
      while ((ch = next ()) != '\n' && ch != EOF)
	;

      if (ch == EOF)
	return "EOF in comment";

      goto COMMENTS;
    }

  if (ch != 'p')
    return "expected 'p cnf ...' header";

  assert (ch == 'p');

  if (fscanf (input, " cnf %d %d", &v, &c) != 2)
    return "invalid header";

  fprintf (output, "c parsed header 'p cnf %d %d'\n", v, c);

  if (v < 0)
    return "negative number of variables";

  max_variable = v;
  NEWN (variables, max_variable + 1l);
  NEWN (variable_order, max_variable + 1l);
  NEWN (variable_elim, max_variable + 1l);
  NEWN (exs, size_exs = max_variable);

  if (c < 0)
    return "negative number of clauses";

  assert (!num_clauses);
  assert (!num_literals);

  lits = 0;

CLAUSES:

  ch = next ();
  if (isspace (ch))
    goto CLAUSES;

  if (ch == EOF)
    {
      if (num_clauses < c)
	return "clause missing";

      assert (num_clauses == c);

      return 0;
    }

  sign = 1;
  if (ch == '-')
    {
      sign = -1;
      ch = next ();
    }

  if (!isdigit (ch))
    return "expected digit";

  lit = 0;

  do
    {
      lit = 10 * lit + (ch - '0');
      ch = next ();
    }
  while (isdigit (ch));

  if (!isspace (ch))
    return "expected white space after literal";

  if (lit > v)
    return "maximal variable index exceeded";

  lit *= sign;

  if (num_literals == size_literals)
    {
      if (size_literals == MAXSIZE) die ("too many literals");
      size_literals *= 2;
      REALLOCN (literals, num_literals, size_literals);
    }

  literals[num_literals++] = lit;

  if (!lit)
    {
      if (num_clauses++ == c)
	return "maximal number of clauses exceeded";

      if (num_clauses == size_clauses)
	{
          if (size_clauses == MAXSIZE) die ("too many clauses");
	  size_clauses *= 2;
	  REALLOCN (clauses, num_clauses, size_clauses);
	}

      clause = clauses + num_clauses;

      clause->lits = lits;
      lits = num_literals;

      num_eles++;

      if (num_eles == size_eles)
	{
          if (size_eles == MAXSIZE) die ("too many 'ELES' in 'parse'");
	  size_eles *= 2;
	  REALLOCN (eles, num_eles, size_eles);
	}

      ele = eles + num_eles;
      ele->type = CLAUSE;
      ele->idx = num_clauses;

    }

  goto CLAUSES;
}

/*------------------------------------------------------------------------*/

static int
clause2chain (Clause * clause)
{
  assert (first_clause_chain);
  return clause - clauses;
}

/*------------------------------------------------------------------------*/

#define pos2wordandbit(words,pos,word,bit) \
  do { \
    assert (pos >= 0); \
 \
    word = pos; \
    bit = (1 << (word & 31)); \
    word >>= 5; \
 \
    assert (word < words[-1]); \
  } while(0)

/*------------------------------------------------------------------------*/

static void
mark (unsigned *words, int idx)
{
  unsigned word, bit;
  pos2wordandbit (words, idx, word, bit);
  words[word] |= bit;
}

/*------------------------------------------------------------------------*/

static int
marked (unsigned *words, int idx)
{
  unsigned word, bit;
  pos2wordandbit (words, idx, word, bit);
  return words[word] & bit;
}

/*------------------------------------------------------------------------*/
#ifndef NDEBUG
/*------------------------------------------------------------------------*/

static int count_bits_in_byte_table[256];
static int count_bits_in_byte_table_initialized;

/*------------------------------------------------------------------------*/

static int
count_bits_in_byte (unsigned char byte)
{
  int i, j, sum;

  if (!count_bits_in_byte_table_initialized)
    {
      for (i = 0; i < 256; i++)
	{
	  sum = 0;
	  for (j = 0; j < 8; j++)
	    if (i & (1 << j))
	      sum++;

	  count_bits_in_byte_table[i] = sum;
	}

      count_bits_in_byte_table_initialized = 1;
    }

  return count_bits_in_byte_table[byte];
}

/*------------------------------------------------------------------------*/

static int
count_marked (unsigned *words, int size)
{
  int res, i, j, tmp;

  res = 0;
  for (i = 0; i < size; i++)
    {
      tmp = words[i];

      for (j = 0; j < 4; j++)
	{
	  res += count_bits_in_byte (tmp);
	  tmp >>= 8;
	}

    }

  return res;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static int
is_constant_bdd_idx (int idx)
{
  return idx == 0 || idx == 1;
}

/*------------------------------------------------------------------------*/

static int
is_trivial_step (int f, int g)
{
  if (is_constant_bdd_idx (f))
    return 1;

  if (is_constant_bdd_idx (g))
    return 1;

  if (f == g)
    return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

static void
check_legal_ex_idx (int idx)
{
#ifndef NDEBUG
  assert (idx >= 0);
  assert (idx < max_variable);
#endif
}

/*------------------------------------------------------------------------*/

static int
ex2bdd (int idx)
{
  check_legal_ex_idx (idx);

  if (idx == TRUE_EX)
    return 1;

  return exs[idx].h;
}

/*------------------------------------------------------------------------*/

static void
check_legal_and_idx (int idx)
{
#ifndef NDEBUG
  if (idx != TRUE_AND)
    {
      if (idx < 0)
	assert (-idx <= num_clauses);
      else
	assert (idx < num_ands);
    }
#endif
}

/*------------------------------------------------------------------------*/

static int
and2bdd (int idx)
{
  check_legal_and_idx (idx);

  if (idx == TRUE_AND)
    return 1;

  return ands[idx].h;
}

/*------------------------------------------------------------------------*/

static int
cl2bdd (int idx)
{
  return clauses[idx].bdd;
}

/*------------------------------------------------------------------------*/

static int
ele2bdd (int i)
{
  Ele *e;

  if (i == TRUE_AND || i == TRUE_EX)
    {
      return 1;
    }

  e = (eles + i);

  switch (e->type)
    {
    case CLAUSE:
      return cl2bdd (e->idx);
    case AND:
      return and2bdd (e->idx);
    case EX:
      return ex2bdd (e->idx);
    default:
      fprintf (output, "i %d, num_eles %d\n", i, num_eles);
      assert (0);
    }
  return 0;
}

/*------------------------------------------------------------------------*/

static int
is_trivial_and (And * and)
{
  int f, g;
  f = ele2bdd (and->a);
  g = ele2bdd (and->b);
  return is_trivial_step (f, g);
}

/*------------------------------------------------------------------------*/

static int
is_redundant_and (And * and)
{
  int f, g, h;

  if (is_trivial_and (and))
    return 0;

  f = ele2bdd (and->a);
  g = ele2bdd (and->b);
  h = and->h;

  return f == h || g == h;
}

/*------------------------------------------------------------------------*/
#ifdef EBDDRES_DUMP
/*------------------------------------------------------------------------*/

void
pretty_print_bdd (const char *mesg, int idx)
{
  Node *n;

  fprintf (output, "BDD: %s", mesg);
  if (idx == 0 || idx == 1)
    {
      fprintf (output, "%d\n", idx);
      return;
    }

  n = (nodes + idx);
  fprintf (output, "var %d\n", n->var);

  fprintf (output, "\tLC ");
  pretty_print_bdd ("", n->child[0]);

  fprintf (output, "\tRC ");
  pretty_print_bdd ("", n->child[1]);
}

/*------------------------------------------------------------------------*/

static void
dump_buckets (const char *mesg)
{
#ifndef NDEBUG
  int i;
  printf ("BC %s:", mesg);
  for (i = 0; i < size_buckets; i++)
    {
      printf (" %d", *(buckets + i));
    }
  printf ("\n");
#endif
}

/*------------------------------------------------------------------------*/

static void
dump_occurrences ()
{
  int i;

  fprintf (output, "OCC:");
  for (i = 0; i <= size_occurrences; i++)
    {
      fprintf (output, " %d", *(occurrences + i));
    }
  fputc ('\n', output);
}

/*------------------------------------------------------------------------*/

static void
dump_and (And * and)
{
  int idx = and - ands;

  fprintf (output, "c and %d: %d & %d = %d", idx, and->a, and->b, and->h);

  if (is_trivial_and (and))
    fputs (" trivial", output);

  if (is_redundant_and (and))
    fputs (" redundant", output);

  if (marked_ands && !marked (marked_ands, idx))
    fputs (" not-in-core", output);

  fputc ('\n', output);
}

/*------------------------------------------------------------------------*/

static void
dump_ele (Ele * e)
{
  fprintf (output, "type %s, index %d ",
	   (e->type == CLAUSE) ? "clause" :
	   (e->type == AND) ? "and" : "ex", e->idx);

  switch (e->type)
    {
    case CLAUSE:
      fprintf (output, "BDD %d\n", (clauses + e->idx)->bdd);
      break;
    case AND:
      fprintf (output, "a %d ", (ands + e->idx)->a);
      fprintf (output, "b %d ", (ands + e->idx)->b);
      fprintf (output, "BDD %d\n", (ands + e->idx)->h);
      break;
    case EX:
      fprintf (output, "a %d ", (exs + e->idx)->a);
      fprintf (output, "BDD %d\n", (exs + e->idx)->h);
      break;
    }
}

/*------------------------------------------------------------------------*/

static void
dump_variable_order ()
{
  int i;
  for (i = 1; i <= max_variable; i++)
    {
      fprintf (output, "%d ", variable_order[i]);
    }
  fputc ('\n', output);
}

/*------------------------------------------------------------------------*/
static void
dump_variable_elim ()
{
  int i;
  for (i = 1; i <= max_variable; i++)
    {
      fprintf (output, "%d ", variable_elim[i]);
    }
  fputc ('\n', output);
}

/*------------------------------------------------------------------------*/

static void
dump_variable (Var * v)
{
  int *p, c;
  fprintf (output, "v variable %ld, ", (long)(v - variables));
  for (p = occurrences + v->occurrences; (c = *p) && (c != -1); p++)
    fprintf (output, " %d", c);
  fputc ('\n', output);
  fprintf (output, "occ %d, cnt %d, bucket %d\n", v->occurrences, v->count,
	   v->bucket);
  return;
}

/*------------------------------------------------------------------------*/

static void
dump_clause (Clause * clause)
{
  int *p, lit;

  fprintf (output, "c clause %ld, ", (long)(clause - clauses));

  for (p = literals + clause->lits; (lit = *p); p++)
    fprintf (output, " %d", lit);

  fprintf (output, " node %d", clause->bdd);

  if (marked_clauses && !marked (marked_clauses, clause - clauses))
    fputs (" not-in-core", output);

  fputc ('\n', output);
}

/*------------------------------------------------------------------------*/
#endif /* EBDDRES_DUMP */
/*------------------------------------------------------------------------*/

static void
check_legal_var_idx (int var)
{
  assert (var > 0);
  assert (var <= max_variable);
}

/*------------------------------------------------------------------------*/

static void
check_legal_ele_idx (int var)
{
  assert (var > 0);
  assert (var <= num_eles);
}

/*------------------------------------------------------------------------*/

static void
check_legal_bdd_idx (int idx)
{
#ifndef NDEBUG
  if (idx < 0)
    printf ("clb called with %d\n", idx);
#endif
  assert (0 <= idx);
  assert (idx < num_nodes);
}

/*------------------------------------------------------------------------*/

static unsigned primes[] = { 111842747, 222924617, 333548309, 446519803 };

/*------------------------------------------------------------------------*/

static int
hash_bdd (int var, int pos, int neg, unsigned salt)
{
  unsigned tmp, i;
  int res;

  assert (sizeof (primes) / sizeof (primes[0]) == 4);

  check_legal_var_idx (var);
  check_legal_bdd_idx (pos);
  check_legal_bdd_idx (neg);

  i = salt;

  tmp = primes[(i++) & 3] * (unsigned) var;
  tmp = primes[(i++) & 3] * (tmp + (unsigned) pos);
  tmp = primes[(i++) & 3] * (tmp + (unsigned) neg);

  tmp &= size_table - 1;
  res = tmp;

  assert (0 <= res);
  assert (res < size_table);

  return res;
}

/*------------------------------------------------------------------------*/

static int
find_bdd (int var, int pos, int neg)
{
  unsigned p, idx, delta;
  Node *node;
  int count;

  assert (pos != neg);

  table_finds++;

  p = hash_bdd (var, pos, neg, 0);
  idx = table[p];
  if (!idx)
    return p;

  node = nodes + idx;
  if (node->var == var && node->child[1] == pos && node->child[0] == neg)
    return p;

  delta = hash_bdd (var, pos, neg, 1);
  if (!(delta & 1))
    delta++;

  assert (delta < size_table);

  count = 1;

  for (;;)
    {
      p += delta;
      if (p >= size_table)
	{
	  p -= size_table;
	  assert (p < size_table);
	}

      idx = table[p];
      if (!idx)
	break;

      node = nodes + idx;
      if (node->var == var && node->child[1] == pos && node->child[0] == neg)
	break;

      count++;
      assert (count < size_table);
    }

  table_collisions += count;

  return p;
}

/*------------------------------------------------------------------------*/

static double
seconds (void)
{
  return current_time () - start_time;
}

/*------------------------------------------------------------------------*/

static double
mb (void)
{
  return max_bytes / (double) (1 << 20);
}

/*------------------------------------------------------------------------*/

static double
progress (void)
{
  if (strategy == 0 || strategy == 1)
    return (num_clauses > 1) ? num_ands / (double) (num_clauses) : 1.0;
  else if (strategy == 2 || strategy == 3 || strategy == 4)
    return (max_variable > 1) ? num_exs / (double) (max_variable) : 1.0;

  return 0.0;
}

/*------------------------------------------------------------------------*/
/**INDENT-OFF**/

static void
report (void)
{
  if (!reports++)
    fprintf (output,
"c\n"
"c  seconds   MB  progress  ands      exs       nodes        steps        lines\n"
"c\n");

  fprintf (output,
	   "c %7.1f %6.1f %6.2f%% %6d %4d/%d %11d %12.0f %12d\n",
	   seconds (), mb (), 100.0 * progress (),
	   num_ands, num_exs, max_variable, num_nodes, num_steps, num_lines);

  last_report_num_steps = num_steps;
  fflush(output);
}

/**INDENT-ON**/
/*------------------------------------------------------------------------*/

static void
enlarge_table (void)
{
  int idx, p;
  Node *node;

  DELETEN (table, size_table);
  if (size_table == MAXSIZE) die ("table too large");
  size_table *= 2;
  NEWN (table, size_table);

  for (idx = 2; idx < num_nodes; idx++)
    {
      check_legal_bdd_idx (idx);
      node = nodes + idx;
      p = find_bdd (node->var, node->child[1], node->child[0]);
      assert (!table[p]);
      table[p] = idx;
    }

  report ();
}

/*------------------------------------------------------------------------*/

static int
cmp_var (int a, int b)
{
  check_legal_var_idx (a);
  check_legal_var_idx (b);
  return variable_order[a] - variable_order[b];
}

/*------------------------------------------------------------------------*/

static int
new_bdd (int var, int pos, int neg)
{
  int p, idx;
  Node *node;

  check_legal_bdd_idx (pos);
  check_legal_bdd_idx (neg);

  assert (nodes[pos].var <= 1 || cmp_var (nodes[pos].var, var) < 0);
  assert (nodes[neg].var <= 1 || cmp_var (nodes[neg].var, var) < 0);

  if (pos == neg)
    return pos;

  if (size_table <= 2 * (num_table + 1))
    enlarge_table ();

  p = find_bdd (var, pos, neg);
  idx = table[p];
  if (idx)
    return idx;

  if (num_nodes >= size_nodes)
    {
      if (size_nodes == MAXSIZE) die ("too many nodes");
      size_nodes *= 2;
      REALLOCN (nodes, num_nodes, size_nodes);
    }

  idx = num_nodes++;
  table[p] = idx;
  num_table++;

  assert (idx < size_nodes);
  node = nodes + idx;
  node->var = var;
  node->child[1] = pos;
  node->child[0] = neg;

  return idx;
}

/*------------------------------------------------------------------------*/

static void
check_legal_cache_idx (int idx)
{
  assert (0 <= idx);
  assert (0 < size_and_cache);
}

/*------------------------------------------------------------------------*/

static void
check_legal_or_cache_idx (int idx)
{
  assert (0 <= idx);
  assert (0 < size_or_cache);
}

/*------------------------------------------------------------------------*/

static int
hash_line (int a, int b, unsigned salt)
{
  unsigned tmp, i;
  int res;

  assert (sizeof (primes) / sizeof (primes[0]) == 4);

  check_legal_bdd_idx (a);
  check_legal_bdd_idx (b);

  i = salt;

  tmp = primes[(i++) & 3] * (unsigned) a;
  tmp = primes[(i++) & 3] * (tmp + (unsigned) b);

  tmp &= size_and_cache - 1;
  res = tmp;

  check_legal_cache_idx (res);

  return res;
}

/*------------------------------------------------------------------------*/
static int
hash_or_line (int a, int b, unsigned salt)
{
  unsigned tmp, i;
  int res;

  assert (sizeof (primes) / sizeof (primes[0]) == 4);

  check_legal_bdd_idx (a);
  check_legal_bdd_idx (b);

  i = salt;

  tmp = primes[(i++) & 3] * (unsigned) a;
  tmp = primes[(i++) & 3] * (tmp + (unsigned) b);

  tmp &= size_or_cache - 1;
  res = tmp;

  check_legal_or_cache_idx (res);

  return res;
}

/*------------------------------------------------------------------------*/

static void
check_non_constant_bdd_idx (int idx)
{
  check_legal_bdd_idx (idx);
  assert (!is_constant_bdd_idx (idx));
}

/*------------------------------------------------------------------------*/
/* Returns the position in the cache in which either the line index for the
 * given arguments is stored, or were it should be stored.  Valid lines have
 * all a positive index, and thus an entry of '0' denotes that the
 * corresponding line has not been generated and inserted yet.
 */
static int
find_and_line (int f, int g)
{
  unsigned p, idx, delta;
  int count;
  Line *line;

  assert (f < g);
  assert (!is_trivial_step (f, g));
  cache_finds++;

  p = hash_line (f, g, 0);
  idx = and_cache[p];
  if (!idx)
    return p;

  line = and_lines + idx;
  if (line->f == f && line->g == g)
    return p;

  count = 1;

  delta = hash_line (f, g, 1);
  if (!(delta & 1))
    delta++;

  assert (delta < size_and_cache);

  for (;;)
    {
      p += delta;
      if (p >= size_and_cache)
	p -= size_and_cache;

      check_legal_cache_idx (p);
      idx = and_cache[p];
      if (!idx)
	break;

      line = and_lines + idx;
      if (line->f == f && line->g == g)
	break;

      count++;
      assert (count < size_and_cache);
    }

  cache_collisions += count;

  return p;
}

/*------------------------------------------------------------------------*/
/* Returns the position in the cache in which either the line index for the
 * given arguments is stored, or were it should be stored.  Valid lines have
 * all a positive index, and thus an entry of '0' denotes that the
 * corresponding line has not been generated and inserted yet.
 */
static int
find_or_line (int f, int g)
{
  unsigned p, idx, delta;
  int count;
  Line *line;

  assert (f < g);
  assert (!is_trivial_step (f, g));
  cache_finds++;

  p = hash_or_line (f, g, 0);
  idx = or_cache[p];
  if (!idx)
    return p;

  line = or_lines + idx;
  if (line->f == f && line->g == g)
    return p;

  count = 1;

  delta = hash_or_line (f, g, 1);
  if (!(delta & 1))
    delta++;

  assert (delta < size_or_cache);

  for (;;)
    {
      p += delta;
      if (p >= size_or_cache)
	p -= size_or_cache;

      check_legal_or_cache_idx (p);
      idx = or_cache[p];
      if (!idx)
	break;

      line = or_lines + idx;
      if (line->f == f && line->g == g)
	break;

      count++;
      assert (count < size_or_cache);
    }

  cache_collisions += count;

  return p;
}

/*------------------------------------------------------------------------*/

static void
enlarge_and_cache (void)
{
  Line *line;
  int i, p;

  DELETEN (and_cache, size_and_cache);
  if (size_and_cache == MAXSIZE) die ("and cache too large");
  size_and_cache *= 2;
  NEWN (and_cache, size_and_cache);

  for (i = 1; i <= num_lines; i++)
    {
      line = and_lines + i;
      p = find_and_line (line->f, line->g);
      assert (!and_cache[p]);
      and_cache[p] = i;
    }

  report ();
}

/*------------------------------------------------------------------------*/

static void
enlarge_or_cache (void)
{
  Line *line;
  int i, p;

  DELETEN (or_cache, size_or_cache);
  if (size_or_cache == MAXSIZE) die ("or cache too large");
  size_or_cache *= 2;
  NEWN (or_cache, size_or_cache);

  for (i = 1; i <= num_or_lines; i++)
    {
      line = or_lines + i;
      p = find_or_line (line->f, line->g);
      assert (!or_cache[p]);
      or_cache[p] = i;
    }

  report ();
}

/*------------------------------------------------------------------------*/

static int
is_redundant_step (int f, int g, int h)
{
  if (is_trivial_step (f, g))
    return 0;

  return f == h || g == h;
}

/*------------------------------------------------------------------------*/

static int
is_redundant_line (Line * line)
{
  return is_redundant_step (line->f, line->g, line->h);
}

/*------------------------------------------------------------------------*/

static void
insert_and_cache (int f, int g, int h)
{
  int p, idx;
  Line *line;

  assert (f < g);
  assert (!is_trivial_step (f, g));
  check_legal_bdd_idx (h);

  if (size_and_cache <= 2 * (num_and_cache + 1))
    enlarge_and_cache ();

  p = find_and_line (f, g);
  assert (!and_cache[p]);

  num_lines++;

  if (num_lines == size_and_lines)
    {
      if (size_and_lines == MAXSIZE) die ("too many and lines");
      size_and_lines *= 2;
      REALLOCN (and_lines, num_lines, size_and_lines);
    }

  idx = num_lines;
  and_cache[p] = idx;
  num_and_cache++;

  assert (idx < size_and_lines);
  line = and_lines + idx;
  line->f = f;
  line->g = g;
  line->h = h;

  if (is_redundant_line (line))
    num_redundant_lines++;
}

/*------------------------------------------------------------------------*/

static void
insert_or_cache (int f, int g, int h)
{
  int p, idx;
  Line *line;

  assert (f < g);
  assert (!is_trivial_step (f, g));
  check_legal_bdd_idx (h);

  if (size_or_cache <= 2 * (num_or_cache + 1))
    enlarge_or_cache ();

  p = find_or_line (f, g);
  assert (!or_cache[p]);

  num_or_lines++;

  if (num_or_lines == size_or_lines)
    {
      if (size_or_lines == MAXSIZE) die ("too many or lines");
      size_or_lines *= 2;
      REALLOCN (or_lines, num_or_lines, size_or_lines);
    }

  idx = num_or_lines;
  or_cache[p] = idx;
  num_or_cache++;

  assert (idx < size_or_lines);
  line = or_lines + idx;
  line->f = f;
  line->g = g;
  line->h = h;

  if (is_redundant_line (line))
    num_redundant_or_lines++;
}

/*------------------------------------------------------------------------*/

static int
is_and_cached (int f, int g, int *h_ptr)
{
  int p, idx;
  Line *line;

  assert (h_ptr);
  assert (f < g);
  assert (!is_trivial_step (f, g));

  cache_lookups++;

  p = find_and_line (f, g);
  idx = and_cache[p];
  if (!idx)
    return 0;

  cache_hits++;

  line = and_lines + idx;
  *h_ptr = line->h;

  return 1;
}

/*------------------------------------------------------------------------*/

static int
is_or_cached (int f, int g, int *h_ptr)
{
  int p, idx;
  Line *line;

  assert (h_ptr);
  assert (f < g);
  assert (!is_trivial_step (f, g));

  cache_lookups++;

  p = find_or_line (f, g);
  idx = or_cache[p];
  if (!idx)
    return 0;

  cache_hits++;

  line = or_lines + idx;
  *h_ptr = line->h;

  return 1;
}

/*------------------------------------------------------------------------*/

#define normalize(f,g) \
  do { \
    if ((f) > (g)) \
      { \
	int tmp = (f); \
	(f) = (g); \
	(g) = tmp; \
      } \
  } while(0)

/*------------------------------------------------------------------------*/

#define cofactor(f, x, f0, f1) \
  do { \
    f0 = (nodes[f].var == x) ? nodes[f].child[0] : f; \
    f1 = (nodes[f].var == x) ? nodes[f].child[1] : f; \
  } while(0)

/*------------------------------------------------------------------------*/

#define cofactors(f, g, x, f0, g0, f1, g1) \
  do { \
    int fvar = nodes[f].var; \
    int gvar = nodes[g].var; \
    x = (cmp_var (fvar, gvar) < 0) ? gvar : fvar; \
    cofactor (f, x, f0, f1); \
    cofactor (g, x, g0, g1); \
  } while (0)

/*------------------------------------------------------------------------*/

static int
and_step (int f, int g)
{
  int h, x, f1, f0, g1, g0, h0, h1;

  check_legal_bdd_idx (f);
  check_legal_bdd_idx (g);

  num_steps++;

  if (f == 0)
    return 0;

  if (f == 1)
    return g;

  if (g == 0)
    return 0;

  if (g == 1)
    return f;

  if (f == g)
    return f;

  /* Number of lines 'num_lines' is thus the number of non trivial steps.
   */
  assert (!is_trivial_step (f, g));

  normalize (f, g);
  if (is_and_cached (f, g, &h))
    return h;

  cofactors (f, g, x, f0, g0, f1, g1);

  h0 = and_step (f0, g0);
  h1 = and_step (f1, g1);

  h = new_bdd (x, h1, h0);

  insert_and_cache (f, g, h);

  return h;
}

/*------------------------------------------------------------------------*/
static int
or_step (int f, int g)
{
  int h, x, f1, f0, g1, g0, h0, h1;

  check_legal_bdd_idx (f);
  check_legal_bdd_idx (g);

  num_steps++;

  if (f == 0)
    return g;

  if (f == 1)
    return 1;

  if (g == 0)
    return f;

  if (g == 1)
    return 1;

  if (f == g)
    return f;


  /* Number of lines 'num_lines' is thus the number of non trivial steps.
   */
  assert (!is_trivial_step (f, g));

  normalize (f, g);
  if (is_or_cached (f, g, &h))
    return h;

  cofactors (f, g, x, f0, g0, f1, g1);

  h0 = or_step (f0, g0);
  h1 = or_step (f1, g1);

  h = new_bdd (x, h1, h0);

  insert_or_cache (f, g, h);

  return h;
}

/*------------------------------------------------------------------------*/

static int
is_trivial_ex (Ex * ex)
{
  int f, g;
  Node *n;
  n = nodes + ele2bdd (ex->a);
  f = n->child[0];
  g = n->child[1];
  return is_trivial_step (f, g);
}

/*------------------------------------------------------------------------*/

static int
is_redundant_ex (Ex * ex)
{
  int f, g, h;
  Node *n;

  if (is_trivial_ex (ex))
    return 0;

  n = nodes + ele2bdd (ex->a);
  f = n->child[0];
  g = n->child[1];
  h = ex->h;

  return f == h || g == h;
}

/*------------------------------------------------------------------------*/

static int
and (int a, int b)
{
  int res, f, g, h;
  And *and;
  Ele *ele;

  f = ele2bdd (a);
  g = ele2bdd (b);

  h = and_step (f, g);

  if (num_ands == size_ands)
    {
      if (size_ands == MAXSIZE) die ("too many ands in 'and'");
      size_ands *= 2;
      REALLOCN (ands, num_ands, size_ands);
    }

  res = num_ands++;
  and = ands + res;
  and->a = a;
  and->b = b;
  and->h = h;

  num_eles++;
  if (num_eles == size_eles)
    {
      if (size_eles == MAXSIZE) die ("too many 'ELES' in 'and'");
      size_eles *= 2;
      REALLOCN (eles, num_eles, size_eles);
    }

  ele = eles + num_eles;
  ele->type = AND;
  ele->idx = res;

  if (is_trivial_and (and))
    num_trivial_ands++;
  else if (is_redundant_and (and))
    num_redundant_ands++;

  return num_eles;
}

/*------------------------------------------------------------------------*/

static int
exists (int a)
{
  int res, f, h;
  Ex *ex;
  Ele *ele;

  f = ele2bdd (a);

  h = or_step (nodes[f].child[0], nodes[f].child[1]);

  res = num_exs++;
  ex = exs + res;
  ex->a = a;
  ex->h = h;

  num_eles++;
  if (num_eles == size_eles)
    {
      if (size_eles == MAXSIZE) die ("too many 'ELES' in 'exists'");
      size_eles *= 2;
      REALLOCN (eles, num_eles, size_eles);
    }

  ele = eles + num_eles;
  ele->type = EX;
  ele->idx = res;

  if (is_trivial_ex (ex))
    num_trivial_exs++;

  if (is_redundant_ex (ex))
    num_redundant_exs++;

  return num_eles;
}

/*------------------------------------------------------------------------*/

static void
check_legal_literals_start (int lits)
{
  assert (0 <= lits);
  assert (lits < num_literals);
  assert (!lits || !literals[lits - 1]);
}

/*------------------------------------------------------------------------*/

static int
length_lits (int *lits)
{
  const int *p = lits;

  while (*p)
    p++;

  return p - lits;
}

/*------------------------------------------------------------------------*/

static int
cmp_lit (const void *p, const void *q)
{
  int k = *(const int *) p;
  int l = *(const int *) q;

  if (l < 0)
    l = -l;

  if (k < 0)
    k = -k;

  return cmp_var (k, l);	/* smaller variables first */
}

/*------------------------------------------------------------------------*/

static int
rev_cmp_lit (const void *p, const void *q)
{
  int k = *(const int *) p;
  int l = *(const int *) q;

  if (l < 0)
    l = -l;

  if (k < 0)
    k = -k;

  return cmp_var (l, k);	/* larger variables first */
}

/*------------------------------------------------------------------------*/

static void
sort_clause (Clause * clause)
{
  int len, *lits;
  check_legal_literals_start (clause->lits);
  lits = literals + clause->lits;
  len = length_lits (lits);
  qsort (lits, len, sizeof (lits[0]), cmp_lit);
}

/*------------------------------------------------------------------------*/

static void
reverse_sort_clause (Clause * clause)
{
  int len, *lits;
  check_legal_literals_start (clause->lits);
  lits = literals + clause->lits;
  len = length_lits (lits);
  qsort (lits, len, sizeof (lits[0]), rev_cmp_lit);
}

/*------------------------------------------------------------------------*/
#ifdef EBDDRES_DUMP
/*------------------------------------------------------------------------*/

static void
forall_ands (void (*f) (And *))
{
  int i;
  for (i = 0; i < num_ands; i++)
    f (ands + i);
}

/*------------------------------------------------------------------------*/

static void
forall_eles (void (*f) (Ele *))
{
  int i;

  for (i = 0; i <= num_eles; i++)
    {
      fprintf (output, "%d: ", i);
      f (eles + i);
    }
}

/*------------------------------------------------------------------------*/

static void
forall_variables (void (*f) (Var *))
{
  int i;

  for (i = 0; i <= max_variable; i++)
    f (variables + i);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static void
forall_clauses (void (*f) (Clause *))
{
  int i;
  assert (num_clauses < size_clauses);
  for (i = 1; i <= num_clauses; i++)
    f (clauses + i);
}

/*------------------------------------------------------------------------*/

static void
generate_bdd_for_clause (Clause * clause)
{
  int *p, lit, node, *lits;

  check_legal_literals_start (clause->lits);
  lits = literals + clause->lits;

  node = 0;
  for (p = lits; (lit = *p); p++)
    node = (lit > 0) ? new_bdd (lit, 1, node) : new_bdd (-lit, node, 1);

  if (!node)
    variables->count = 1;	/* ugly */

  clause->bdd = node;

  if (!empty_clause && !node)
    empty_clause = clause;
}

/*------------------------------------------------------------------------*/

static void
forall_bdds (void (*f) (Node *))
{
  int i;
  for (i = 2; i < num_nodes; i++)
    f (nodes + i);
}

/*------------------------------------------------------------------------*/

static void
forall_lines (void (*f) (Line *))
{
  int i;
  for (i = 1; i <= num_lines; i++)
    f (and_lines + i);
}

/*------------------------------------------------------------------------*/

static void
forall_or_lines (void (*f) (Line *))
{
  int i;
  for (i = 1; i <= num_or_lines; i++)
    f (or_lines + i);
}

/*------------------------------------------------------------------------*/

static void
forall_exs (void (*f) (Ex *))
{
  int i;
  for (i = 0; i < num_exs; i++)
    f (exs + i);
}

/*------------------------------------------------------------------------*/

static void
init_count (void)
{
  int i, idx;

  for (i = 0; i < num_literals; i++)
    {
      idx = literals[i];
      if (!idx)
	continue;

      if (idx < 0)
	idx = -idx;

      check_legal_var_idx (idx);
      variables[idx].count++;
    }
}

/*------------------------------------------------------------------------*/

static void
check_legal_clause_idx (int idx)
{
  assert (0 < idx);
  assert (idx <= num_clauses);
}

/*------------------------------------------------------------------------*/

static void
init_occurrences (void)
{
  int i, j, *p, lit;
  Clause *clause;
  Var *v;

  init_count ();

  assert (num_literals >= num_clauses);

  size_occurrences = num_literals;
  size_occurrences -= num_clauses;	/* ignore 0 at end of clause */
  size_occurrences += max_variable;	/* one -1 sentinel per variable */

  NEWN (occurrences, size_occurrences);

  j = 0;
  for (i = 1; i <= max_variable; i++)
    {
      assert (!variables[i].occurrences);
      variables[i].occurrences = j;
      j += variables[i].count;
      assert (j < size_occurrences);
      occurrences[j++] = -1;	/* sentinel */
    }

  assert (j == size_occurrences);

  for (i = 1; i <= num_clauses; i++)
    {
      check_legal_clause_idx (i);
      clause = clauses + i;
      check_legal_literals_start (clause->lits);

      for (p = literals + clause->lits; (lit = *p); p++)
	{
	  if (lit < 0)
	    lit = -lit;

	  check_legal_var_idx (lit);
	  v = variables + lit;
	  assert (!occurrences[v->occurrences]);
	  occurrences[v->occurrences++] = i;
	}

    }

  for (i = 1; i <= max_variable; i++)
    {
      check_legal_var_idx (i);
      v = variables + i;
      assert (occurrences[v->occurrences] == -1);
      while (v->occurrences > 0 && occurrences[v->occurrences - 1] >= 0)
	v->occurrences--;
      assert (occurrences[v->occurrences + v->count] == -1);
    }
}

/*------------------------------------------------------------------------*/

static void
init_original_order (void)
{
  int i;

  fprintf (output, "c original variable order\n");
  for (i = 1; i <= max_variable; i++)
    variable_order[i] = i;
}

/*------------------------------------------------------------------------*/

static void
init_reverse_order (void)
{
  int i;
  fprintf (output, "c reverse original variable order\n");
  for (i = 1; i <= max_variable; i++)
    variable_order[i] = max_variable - i + 1;
}

/*------------------------------------------------------------------------*/

static int
cmp_count (const void *p, const void *q)
{
  int a = *(const int *) p;
  int b = *(const int *) q;
  int res;

  if (a < 0)
    a = -a;

  if (b < 0)
    b = -b;

  check_legal_var_idx (a);
  check_legal_var_idx (b);

  res = variables[a].count - variables[b].count;
  if (res)
    return res;

  return variable_order[a] - variable_order[b];
}

/*------------------------------------------------------------------------*/

static int
reverse_cmp_count (const void *p, const void *q)
{
  return -cmp_count (p, q);
}

/*------------------------------------------------------------------------*/

static void
sort_clause_by_count (Clause * clause)
{
  int len, *lits;
  check_legal_literals_start (clause->lits);
  lits = literals + clause->lits;
  len = length_lits (lits);
  qsort (lits, len, sizeof (lits[0]), reverse_cmp_count);
}

/*------------------------------------------------------------------------*/

static void
push_all_variables (void)
{
  int idx;

  assert (!num_stack);
  for (idx = 1; idx <= max_variable; idx++)
    push (idx);
}

/*------------------------------------------------------------------------*/

static void
forall_variables_in_clauses_with (int this, void (*f) (int, int))
{
  int *p, *q, other, clause_idx;
  Clause *clause;

  for (p = occurrences + variables[this].occurrences;
       (clause_idx = *p) >= 0; p++)
    {
      check_legal_clause_idx (clause_idx);
      clause = clauses + clause_idx;
      check_legal_literals_start (clause->lits);

      for (q = literals + clause->lits; (other = *q); q++)
	{
	  if (other < 0)
	    other = -other;

	  check_legal_var_idx (other);
	  f (this, other);
	}
    }
}

/*------------------------------------------------------------------------*/

static void
push_other (int this, int other)
{
  push (other);
}

/*------------------------------------------------------------------------*/

static void
init_dfs_order (void)
{
  int idx, pos;

  fprintf (output, "c DFS variable order\n");

  if (!max_variable)
    return;

  push_all_variables ();

  qsort (stack, num_stack, sizeof (stack[0]), reverse_cmp_count);

  forall_clauses (sort_clause_by_count);

  pos = 1;

  while (num_stack > 0)
    {
      idx = pop ();
      check_legal_var_idx (idx);

      if (variable_order[idx])
	continue;

      assert (pos);
      variable_order[idx] = pos++;

      forall_variables_in_clauses_with (idx, push_other);
    }

  assert (pos == max_variable + 1);
}

/*------------------------------------------------------------------------*/

static void
init_sum_order (void)
{
  int i, pos;

  fprintf (output, "c sum variable order\n");

  push_all_variables ();
  qsort (stack, num_stack, sizeof (stack[0]), cmp_count);
  pos = 1;
  for (i = 0; i < num_stack; i++)
    variable_order[stack[i]] = pos++;
  assert (pos == max_variable + 1);
  num_stack = 0;
}

/*------------------------------------------------------------------------*/

static int optimize_order = 1;
static InitOrder init_order_function = init_original_order;

/*------------------------------------------------------------------------*/

static void
init_sorted (void)
{
  int i;
  assert (!num_stack);
  for (i = 1; i <= max_variable; i++)
    push (i);
  qsort (stack, num_stack, sizeof (stack[0]), cmp_lit);
}

/*------------------------------------------------------------------------*/

static void
dump_order (void)
{
  int i;
  assert (order_file);
  init_sorted ();
  for (i = 0; i < num_stack; i++)
    fprintf (order_file, "%d\n", stack[i]);
  num_stack = 0;
  fflush (order_file);
}

/*------------------------------------------------------------------------*/

static void
check_order (void)
{
#ifndef NDEBUG
  int i;
  init_sorted ();
  for (i = 0; i < num_stack - 1; i++)
    assert (variable_order[stack[i]] < variable_order[stack[i + 1]]);
  num_stack = 0;
#endif
}

/*------------------------------------------------------------------------*/

static void
inc_width (int this, int other)
{
  if (variable_order[this] >= variable_order[other])
    return;

  variable_widths[other]++;
  variable_cut++;
}

/*------------------------------------------------------------------------*/

static void
clr_widths (void)
{
  memset (variable_widths, 0,
	  sizeof (variable_widths[0]) * (max_variable + 1));
}

/*------------------------------------------------------------------------*/

static int64
cut_width (void)
{
  int64 sum_widths, max_width, tmp;
  int i, this;

  if (variable_widths)
    clr_widths ();
  else
    NEWN (variable_widths, max_variable + 1);

  init_sorted ();
  variable_cut = 0;
  for (i = 0; i < num_stack; i++)
    {
      this = stack[i];
      forall_variables_in_clauses_with (this, inc_width);
      assert (variable_cut >= variable_widths[this]);
      variable_cut -= variable_widths[this];
      variable_widths[this] = variable_cut;
    }
  assert (variable_cut == 0);

  sum_widths = 0;
  max_width = 0;

  for (i = 1; i <= max_variable; i++)
    {
      tmp = variable_widths[i];
      if (tmp > max_width)
	max_width = tmp;

      sum_widths += tmp;
    }

  num_stack = 0;

  return sum_widths;
}

/*------------------------------------------------------------------------*/

static void
copy_order (int *dst, int *src)
{
  memcpy (dst, src, sizeof (src[0]) * (max_variable + 1));
}

/*------------------------------------------------------------------------*/

static void
sum_up_order (int this, int other)
{
  check_legal_var_idx (this);
  variable_widths[this] += variable_order[other];
}

/*------------------------------------------------------------------------*/

static int
cmp_width (const void *p, const void *q)
{
  int a = *(const int *) p;
  int b = *(const int *) q;
  int64 aw = variable_widths[a];
  int64 bw = variable_widths[b];
  int res;

  if (aw < bw)
    return -1;

  if (aw > bw)
    return 1;

  res = variables[a].count - variables[b].count;
  if (res)
    return res;

  return a - b;
}

/*------------------------------------------------------------------------*/

static int64
force_order_iteration (int64 prev_cut_width)
{
  int *prev, idx, i;
  int64 res;
  Var *v;

  NEWN (prev, max_variable + 1l);
  copy_order (prev, variable_order);
  clr_widths ();

  for (idx = 1; idx <= max_variable; idx++)
    {
      forall_variables_in_clauses_with (idx, sum_up_order);
      v = variables + idx;
      if (!v->count)
	continue;

      variable_widths[idx] *= num_clauses;
      variable_widths[idx] /= v->count;
    }

  push_all_variables ();
  qsort (stack, num_stack, sizeof (stack[0]), cmp_width);
  for (i = 0; i < num_stack; i++)
    variable_order[stack[i]] = i + 1;
  num_stack = 0;
  check_order ();

  res = cut_width ();
  if (res >= prev_cut_width)
    {
      copy_order (variable_order, prev);
      res = 0;
    }

  DELETEN (prev, max_variable + 1l);

  return res;
}

/*------------------------------------------------------------------------*/

static void
force_order (void)
{
  int iteration;
  int64 cw;

  if (!max_variable)
    return;

  iteration = 0;
  cw = cut_width ();

  do
    {
      fprintf (output, "c average cut width %.3f\n",
	       ((double) cw) / (double) max_variable);

      if (!optimize_order)
	break;

      cw = force_order_iteration (cw);
      iteration++;
    }
  while (cw);

  fprintf (output, "c %d optimizations of variable order\n", iteration);
}

/*------------------------------------------------------------------------*/

static int
cmp_clause (const void *p, const void *q)
{
  int a = *(const int *) p;
  int b = *(const int *) q;
  int64 e, f;

  check_legal_clause_idx (a);
  check_legal_clause_idx (b);

  e = clause_pos[a];
  f = clause_pos[b];

  if (e < f)
    return -1;

  if (e > f)
    return 1;

  return a - b;
}

/*------------------------------------------------------------------------*/

static void
schedule_clauses (void)
{
  int i, j, *p, lit, count;
  Clause *clause;
  int64 pos;

  NEWN (schedule, num_clauses);
  if (orig_schedule)
    {
      j = 0;
      for (i = 1; i <= num_clauses; i++)
	schedule[j++] = i;
    }
  else
    {
      NEWN (clause_pos, num_clauses + 1l);

      j = 0;
      for (i = 1; i <= num_clauses; i++)
	{
	  pos = 0;
	  clause = clauses + i;
	  count = 0;

	  for (p = literals + clause->lits; (lit = *p); p++)
	    {
	      if (lit < 0)
		lit = -lit;

	      pos += variable_order[lit];
	      count++;
	    }

	  pos *= max_variable;
	  if (count > 0)
	    pos /= count;

	  clause_pos[i] = pos;
	  schedule[j++] = i;
	}
      assert (j == num_clauses);
      qsort (schedule, num_clauses, sizeof (schedule[0]), cmp_clause);
    }
}

/*------------------------------------------------------------------------*/

static void
init_order (void)
{
  init_occurrences ();
  init_order_function ();
  force_order ();
  check_order ();
  if (order_file)
    dump_order ();

  schedule_clauses ();
}

/*------------------------------------------------------------------------*/

static void
linear_computation (void)
{
  int i;
  for (i = 0; i < num_clauses; i++)
    accu = and (accu, schedule[i]);
}

/*------------------------------------------------------------------------*/

static void
tree_computation (void)
{
  int i, n;

  n = num_clauses;

  while (n > 1)
    {
      for (i = 0; i < n; i += 2)
	{
	  if (i + 1 < n)
	    accu = and (schedule[i], schedule[i + 1]);
	  else
	    accu = schedule[i];

	  schedule[i / 2] = accu;
	}

      n = (n + 1) / 2;
    }
}

/*------------------------------------------------------------------------*/

static void
init_buckets ()
{
  int i, j, occ, iter, lit;
  Var *v;
  Clause *cl;
  /* each clause once + sentinel */
  size_buckets = num_clauses + max_variable;
  NEWN (buckets, size_buckets);

  forall_clauses (reverse_sort_clause);

  iter = 0;
  for (i = max_variable; i >= 1; i--)
    {
      for (j = 1; i <= max_variable; j++)
	{
	  if (variable_order[j] == i)
	    {
	      variable_elim[max_variable - i + 1] = j;
	      break;
	    }
	}

      v = variables + j;
      v->bucket = iter;
      v->bcnt = 0;
      occ = variables[j].occurrences;
      while (occurrences[occ] != -1)
	{
	  cl = clauses + occurrences[occ];
	  lit = literals[cl->lits];
	  lit = lit > 0 ? lit : -lit;

	  if (lit == j)
	    {
	      buckets[iter++] = occurrences[occ];
	      v->bcnt++;
	    }

	  occ = occ + 1;
	}
      buckets[iter++] = -1;
    }

  forall_clauses (sort_clause);
}

/*------------------------------------------------------------------------*/

static void
adjust_bucket (int cnt, int new_ele)
{
  int i, cum_bucket, insert, n_var;
  Var *nv;

  /* next variable */
  n_var = nodes[ele2bdd (new_ele)].var;
  insert = (n_var != -1) ? (variables + n_var)->bucket - 1 : -1;

  if (insert != -1)
    {
      *(buckets + insert) = new_ele;
      (variables + n_var)->bcnt++;
      (variables + n_var)->bucket--;
    }


  cum_bucket = 0;
  for (i = cnt; i <= max_variable; i++)
    {
      nv = variables + variable_elim[i];
      memcpy (buckets + cum_bucket, buckets + nv->bucket,
	      nv->bcnt * sizeof (int));
      nv->bucket = cum_bucket;
      cum_bucket += nv->bcnt;
      *(buckets + cum_bucket) = -1;
      cum_bucket++;
    }

}



/*------------------------------------------------------------------------*/
/* Try to compute conjcuntionf of several buckets and eliminate multiple
 * variables at the same time:
 *
 * - store last BDD index
 * - if after first AND below threshold, take next var
 */
static int
ve_trial ()
{
  int i = 1, j, n;
  int *p, bcnt, cl, cumu = TRUE_AND, cnt;
  int accum_nodes, prev;
  int temp[trial_tresh];
  Var *v;

  temp[0] = TRUE_AND;
  while (i <= max_variable)
    {
      cnt = 0;
      accum_nodes = 0;
      while (accum_nodes < elim_tresh)
	{
	  cumu = TRUE_AND;
	  prev = num_nodes;

	  if (i + cnt > max_variable)
	    break;
	  v = (variables + variable_elim[i + cnt]);
	  bcnt = v->bcnt;

	  for (p = buckets + v->bucket, j = 0; (cl = *p) && j < bcnt;
	       p++, j++)
	    cumu = and (cumu, cl);

	  temp[cnt] = cumu;

	  accum_nodes += num_nodes - prev;
	  cnt++;
	}

      cumu = temp[0];
      n = cnt;
      while (n > 1)
	{
	  for (j = 0; j < n; j += 2)
	    {
	      if (j + 1 < cnt)
		cumu = and (temp[j], temp[j + 1]);
	      else
		cumu = temp[j];

	      temp[j / 2] = cumu;
	    }
	  n = (n + 1) / 2;
	}

      for (j = 0; j < cnt && (i + j) < max_variable; j++)
	if ((nodes + ele2bdd (cumu))->var == variable_elim[i + j])
	  cumu = exists (cumu);

      if (!ele2bdd (cumu))
	return cumu;

      i += cnt;
      if (i < max_variable)
	adjust_bucket (i, cumu);

    }

  return cumu;
}

/*------------------------------------------------------------------------*/

static int
ve (int var)
{
  int *p, i, bcnt, cl, cumu;

  cumu = TRUE_AND;

  bcnt = (variables + var)->bcnt;

  for (p = buckets, i = 0; (cl = *p) && i < bcnt; p++, i++)
    cumu = and (cumu, cl);

  if ((nodes + ele2bdd (cumu))->var == var)
    cumu = exists (cumu);

  return cumu;
}

/*------------------------------------------------------------------------*/

static int
ve_tree (int var)
{
  int i, bcnt, cumu = TRUE_AND;

  bcnt = (variables + var)->bcnt;

  assert ((variables + var)->bucket == 0);

  if (bcnt == 1)
    cumu = buckets[0];

  while (bcnt > 1)
    {
      for (i = 0; i < bcnt; i += 2)
	{
	  if (i + 1 < bcnt)
	    cumu = and (buckets[i], buckets[i + 1]);
	  else
	    cumu = buckets[i];

	  buckets[i / 2] = cumu;
	}

      bcnt = (bcnt + 1) / 2;
    }


  if ((nodes + ele2bdd (cumu))->var == var)
    cumu = exists (cumu);

  return cumu;
}

/*------------------------------------------------------------------------*/

static void
variable_elimination_tresh ()
{

  /* do test for empty clauses */
  if (variables->count == 1)
    {
      accu = 0;
      return;
    }

  accu = ve_trial ();
  return;
}

/*------------------------------------------------------------------------*/

static void
variable_elimination ()
{
  int i;

  /* do test for empty clauses */
  if (variables->count == 1)
    {
      accu = 0;
      return;
    }

  for (i = 1; i <= max_variable; i++)
    {
      if (strategy == 2)
	accu = ve (variable_elim[i]);
      else if (strategy == 3)
	accu = ve_tree (variable_elim[i]);

      if (!ele2bdd (accu))
	return;

      adjust_bucket (i + 1, accu);
    }
}

/*------------------------------------------------------------------------*/

static int
sat (void)
{
  double start_sat_time = current_time ();

  init_order ();

  forall_clauses (sort_clause);
  forall_clauses (generate_bdd_for_clause);

  accu = TRUE_AND;
  switch (strategy)
    {
    case 0:
      linear_computation ();
      break;
    case 1:
      tree_computation ();
      break;
    case 2:
    case 3:
      init_buckets ();
      variable_elimination ();
      break;
    case 4:
      init_buckets ();
      variable_elimination_tresh ();
    }

  if (reports)
    {
      if (last_report_num_steps != num_steps)
	{
	  report ();
	}

      fputs ("c\n", output);
    }

  sat_time = current_time () - start_sat_time;
  sat_time = (sat_time < 0) ? 0 : sat_time;
  fprintf (output, "c solved instance in %.2f seconds\n", sat_time);

  return ele2bdd (accu) ? SATISFIABLE : UNSATISFIABLE;
}

/*------------------------------------------------------------------------*/

static void
extract_model (void)
{
  int idx, child;
  Node *node;

  if (accu == TRUE_AND)
    return;

  for (idx = ele2bdd (accu); idx != 1; idx = child)
    {
      node = nodes + idx;
      child = node->child[0];
      if (child)
	continue;

      variables[node->var].assignment = 1;
      child = node->child[1];
    }
}

/*------------------------------------------------------------------------*/

static int
find_path_asgn (int node)
{
  Node *n;

  while (node != 0 && node != 1)
    {
      n = nodes + node;
      if (!variables[n->var].assignment)
	node = n->child[0];
      else
	node = n->child[1];
    }

  assert (node == 0 || node == 1);
  return node;
}

/*------------------------------------------------------------------------*/

static void
extract_model_ve (void)
{
  int i;
  Ex *ex;
  Node *n;

  for (i = num_exs - 1; i >= 0; i--)
    {
      ex = exs + i;
      n = nodes + ele2bdd (ex->a);

      if (n->child[0] == 1 || n->child[1] == 0)
	continue;

      if (n->child[0] == 0 || n->child[1] == 1
	  || find_path_asgn (n->child[1]))
	variables[n->var].assignment = 1;
    }
}

/*------------------------------------------------------------------------*/

static void
flush_buffer (void)
{
  assert (num_buffer < sizeof (buffer));
  buffer[num_buffer] = 0;
  fputs (buffer, output);
  fputc ('\n', output);
  num_buffer = 0;
  buffer[num_buffer] = 0;
}

/*------------------------------------------------------------------------*/

static void
print_int_2_buffer (int d)
{
  static char str[20];
  int len;

  sprintf (str, " %d", d);
  len = strlen (str);
  if (num_buffer + len > 79)
    flush_buffer ();

  if (!num_buffer)
    {
      buffer[0] = 'v';
      num_buffer = 1;
      buffer[num_buffer] = 0;
    }

  strcat (buffer + num_buffer, str);
  num_buffer += len;
}

/*------------------------------------------------------------------------*/

static void
print_assignment (void)
{
  int i;

  for (i = 1; i <= max_variable; i++)
    print_int_2_buffer ((variables[i].assignment ? 1 : -1) * i);

  print_int_2_buffer (0);

  if (num_buffer)
    flush_buffer ();
}

/*------------------------------------------------------------------------*/

static void
puti (int n, FILE * file)
{
#if 1
  int tmp = n;
  char *p;

  if (tmp)
    {
      if (tmp < 0)
	{
	  fputc ('-', file);
	  tmp = -tmp;
	}

      p = buffer + sizeof (buffer);
      *--p = 0;

      while (tmp)
	{
	  *--p = '0' + (tmp % 10);
	  tmp /= 10;
	}

      fputs (p, file);
    }
  else
    fputc ('0', file);
#else
  fprintf (file, "%d", n);
#endif
}

/*------------------------------------------------------------------------*/

static void
putispace (int n, FILE * file)
{
  puti (n, file);
  fputc (' ', file);
}

/*------------------------------------------------------------------------*/

static void
put2i (int a, int b, FILE * file)
{
  puti (a, file);
  fputc (' ', file);
  puti (b, file);
}

/*------------------------------------------------------------------------*/

static void
put2ispace (int a, int b, FILE * file)
{
  put2i (a, b, file);
  fputc (' ', file);
}

/*------------------------------------------------------------------------*/

static void
put3i (int a, int b, int c, FILE * file)
{
  put2i (a, b, file);
  fputc (' ', file);
  puti (c, file);
}

/*------------------------------------------------------------------------*/

static void
put3ispace (int a, int b, int c, FILE * file)
{
  put3i (a, b, c, file);
  fputc (' ', file);
}

/*------------------------------------------------------------------------*/

static void
put4i (int a, int b, int c, int d, FILE * file)
{
  put3i (a, b, c, file);
  fputc (' ', file);
  puti (d, file);
}

/*------------------------------------------------------------------------*/

static void
put4ispace (int a, int b, int c, int d, FILE * file)
{
  put4i (a, b, c, d, file);
  fputc (' ', file);
}

/*------------------------------------------------------------------------*/

static void
put5i (int a, int b, int c, int d, int e, FILE * file)
{
  puti (a, file);
  fputc (' ', file);
  puti (b, file);
  fputc (' ', file);
  puti (c, file);
  fputc (' ', file);
  puti (d, file);
  fputc (' ', file);
  puti (e, file);
}

/*------------------------------------------------------------------------*/

static void
put5ispace (int a, int b, int c, int d, int e, FILE * file)
{
  put5i (a, b, c, d, e, file);
  fputc (' ', file);
}

/*------------------------------------------------------------------------*/

static void
print_original_clause_as_chain (Clause * clause, FILE * file)
{
  int *p;

  putispace (clause - clauses, file);

  for (p = literals + clause->lits; *p; p++)
    putispace (*p, file);

  fputs ("0 0\n", file);
}

/*------------------------------------------------------------------------*/

static void
trace_clause (Clause * clause)
{
  traced_chains++;

  if (!marked (marked_clauses, clause - clauses))
    return;

  assert (trace_file);
  assert (traced_chains == clause2chain (clause));

  print_original_clause_as_chain (clause, trace_file);
}

/*------------------------------------------------------------------------*/

static void
trace_ternary_defined_clause (int chain, int a, int b, int c)
{
  put4i (chain, a, b, c, trace_file);
  fputs (" 0 0\n", trace_file);
}

/*------------------------------------------------------------------------*/
/* Returns the variable introduced for this BDD node with the given index.
 */
static int
node2def (int idx)
{
  int res;

  check_legal_bdd_idx (idx);

  if (idx == 0)
    return -true_variable;

  if (idx == 1)
    return true_variable;

  assert (idx >= 2);
  res = first_node_variable + (idx - 2);

  return res;
}

/*------------------------------------------------------------------------*/
/* Returns the index of the chain representing one of the clauses in the
 * definition of the variable 'node2def(idx)'.  The definition itself is
 * an 'if-then-else'.  Thus four clauses are needed.  The function returns
 * the index of the chain where 'node2def(idx)' occurs with 'sign' and the
 * variable of the BDD, e.g. the condition of the 'if-then-else', occurs
 * with 'var_sign'.
 */
static int
node2chain (int idx, int sign, int var_sign)
{
  int res;

  assert (first_node_chain);
  check_non_constant_bdd_idx (idx);
  assert (sign == -1 || sign == 1);
  assert (var_sign == -1 || var_sign == 1);

  assert (idx < num_nodes);
  res = (idx - 2) * 4 + first_node_chain;

  if (sign > 0)
    res += 2;

  if (var_sign > 0)
    res++;

  return res;
}

/*------------------------------------------------------------------------*/
#ifdef EBDDRES_DUMP
/*------------------------------------------------------------------------*/

static void
dump_node (Node * n)
{
  int idx = n - nodes;
  fprintf (output,
	   "c node %d: %d ? %d : %d", idx, n->var, n->child[1], n->child[0]);

  if (first_node_chain)
    fprintf (output, " chain %d %d %d %d",
	     node2chain (idx, -1, -1),
	     node2chain (idx, -1, 1),
	     node2chain (idx, 1, -1), node2chain (idx, 1, 1));

  if (marked_nodes && !marked (marked_nodes, idx))
    fputs (" not-in-core", output);

  fputc ('\n', output);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static void
trace_node (Node * node)
{
  int def, var, idx, pos, neg;
  int chain0 = ++traced_chains;
  int chain1 = ++traced_chains;
  int chain2 = ++traced_chains;
  int chain3 = ++traced_chains;

  ++traced_variables;

  idx = node - nodes;
  if (!no_core && !marked (marked_nodes, idx))
    return;

  def = node2def (idx);
  assert (traced_variables == def);

  var = node->var;
  pos = node2def (node->child[1]);
  neg = node2def (node->child[0]);

  trace_ternary_defined_clause (chain0, -def, -var, pos);
  assert (chain0 == node2chain (idx, -1, -1));

  trace_ternary_defined_clause (chain1, -def, var, neg);
  assert (chain1 == node2chain (idx, -1, 1));

  trace_ternary_defined_clause (chain2, def, -var, -pos);
  assert (chain2 == node2chain (idx, 1, -1));

  trace_ternary_defined_clause (chain3, def, var, -neg);
  assert (chain3 == node2chain (idx, 1, 1));

  if (extended_trace_format)
    fputc ('\n', trace_file);
}

/*------------------------------------------------------------------------*/

static int
clause2unit (int idx)
{
  check_legal_clause_idx (idx);
  return first_root_unit_chain + idx - 1;
}

/*------------------------------------------------------------------------*/
/* Produce a chain for each original clause.  The chain is a unit clause and
 * proves that the BDD build for 'clause' is true.
 */
static void
trace_root_unit_clause (Clause * clause)
{
  int idx, child, var_sign;
  Node *node;

  traced_chains++;
  assert (traced_chains == clause2unit (clause - clauses));

  if (!marked (marked_clauses, clause - clauses))
    return;

  putispace (traced_chains, trace_file);
  if (extended_trace_format)
    put2ispace (node2def (clause->bdd), 0, trace_file);
  else
    fputs ("* ", trace_file);

  putispace (clause2chain (clause), trace_file);

  idx = clause->bdd;
  while (idx)
    {
      check_legal_bdd_idx (idx);
      node = nodes + idx;

      if (node->child[1] == 1)
	{
	  child = node->child[0];
	  var_sign = -1;
	}
      else
	{
	  assert (node->child[0] == 1);

	  child = node->child[1];
	  var_sign = 1;
	}

      putispace (node2chain (idx, 1, var_sign), trace_file);
      if (child)
	putispace (node2chain (idx, 1, -var_sign), trace_file);

      idx = child;
    }

  puti (true_chain, trace_file);
  fputs (" 0\n", trace_file);
}

/*------------------------------------------------------------------------*/

static void
add_node_to_core (int idx)
{
  if (is_constant_bdd_idx (idx))
    return;

  if (marked (marked_nodes, idx))
    return;

  mark (marked_nodes, idx);
  num_core_nodes++;
}

/*------------------------------------------------------------------------*/

static void
add_clause_to_core (int idx)
{
  check_legal_clause_idx (idx);


  if (marked (marked_clauses, idx))
    return;

  add_node_to_core (clauses[idx].bdd);
  mark (marked_clauses, idx);
  num_core_clauses++;
}

/*------------------------------------------------------------------------*/

static void
check_legal_line_idx (int idx)
{
  assert (0 < idx);
  assert (idx <= num_lines);
}

/*------------------------------------------------------------------------*/

static void
check_legal_or_line_idx (int idx)
{
  assert (0 < idx);
  assert (idx <= num_or_lines);
}

/*------------------------------------------------------------------------*/

static int
step2idx (int f, int g)
{
  int p, idx;
  normalize (f, g);
  p = find_and_line (f, g);
  idx = and_cache[p];
  assert (idx);
  return idx;
}

/*------------------------------------------------------------------------*/

static int
or_step2idx (int f, int g)
{
  int p, idx;
  normalize (f, g);
  p = find_or_line (f, g);
  idx = or_cache[p];
  assert (idx);
  return idx;
}

/*------------------------------------------------------------------------*/

static void
add_and_line_to_core (int idx)
{
  check_legal_line_idx (idx);
  if (marked (marked_and_lines, idx))
    return;

  mark (marked_and_lines, idx);
  num_core_lines++;
}

/*------------------------------------------------------------------------*/

static void
add_or_line_to_core (int idx)
{
  check_legal_or_line_idx (idx);
  if (marked (marked_or_lines, idx))
    return;

  mark (marked_or_lines, idx);
  num_core_or_lines++;
}

/*------------------------------------------------------------------------*/

static void
add_ex_to_core (int idx)
{
  Node *n;
  assert (idx >= 0);
  assert (idx < max_variable);

  if (marked (marked_exs, idx))
    return;

  n = nodes + ele2bdd (exs[idx].a);

  if (is_trivial_step (n->child[0], n->child[1]))
    num_trivial_core_exs++;
  else
    add_or_line_to_core (or_step2idx (n->child[0], n->child[1]));

  mark (marked_exs, idx);
  num_core_exs++;
}

/*------------------------------------------------------------------------*/

static void
add_non_clause_and_to_core (int idx)
{
  And *and;
  int f, g;


  assert (idx >= 0);
  assert (idx < num_ands);

  if (marked (marked_ands, idx))
    return;

  mark (marked_ands, idx);
  num_core_ands++;

  and = ands + idx;

  if (is_trivial_and (and))
    num_trivial_core_ands++;
  else if (is_redundant_and (and))
    num_redundant_core_ands++;
  else
    {
      f = ele2bdd (and->a);
      g = ele2bdd (and->b);
      add_and_line_to_core (step2idx (f, g));
    }
}

/*------------------------------------------------------------------------*/

static void
add_ele_to_core (int idx)
{
  Ele *e;
  check_legal_ele_idx (idx);

  if (marked (marked_eles, idx))
    return;

  mark (marked_eles, idx);


  e = eles + idx;
  switch (e->type)
    {
    case CLAUSE:
      add_clause_to_core (e->idx);
      break;
    case AND:
      add_non_clause_and_to_core (e->idx);
      break;
    case EX:
      add_ex_to_core (e->idx);
      break;
    default:
      assert (0);
    }
}

/*------------------------------------------------------------------------*/
/* First try to find a root for the resolution proof, which is supposed to
 * be the index of the first 'Ele' that generated a false BDD, or if the CNF
 * is satisfiable, the index of the last 'Ele'.  If no 'Ele' was generated,
 * then the result is 'TRUE_AND' and the function returns '0', otherwise
 * '1'.  Ultimately, after finishing 'core_ands' the 'root' will be made non
 * redundant, but 'find_root' may produce a redundant 'root'.
 */
static int
find_root (void)
{
  Ele *e;
  if (num_eles > 0)
    {
      for (root = 1; root < num_eles; root++)
	if (!ele2bdd (root))
	  break;
    }
  else
    root = TRUE_AND;

  /* satisfiable case, check if root is exists (variable elimination) */
  if (root != TRUE_AND && ele2bdd (root) == 1)
    {
      e = eles + root;
      if (e->type == EX)
	root = (exs + e->idx)->a;
    }

  return root != TRUE_AND;
}

/*------------------------------------------------------------------------*/

#define PERCENT(a,b) ((a) ? ((a) * 100.0) / (double) (b) : 0.0)
#define RELATIVE(a,b) ((a) ? (a) / (double) (b) : 0.0)

/*------------------------------------------------------------------------*/

static void
single_child (Ex * ex)
{
  add_ele_to_core (ex->a);
  return;
}

/*------------------------------------------------------------------------*/

static void
two_children (And * and)
{
  int a, b, f, g, h;

#ifndef NDEBUG
  int redundant, trivial;
#endif

  a = and->a;
  b = and->b;
  h = and->h;
  f = ele2bdd (a);
  g = ele2bdd (b);

#ifndef NDEBUG
  redundant = 0;
  trivial = 1;
#endif
  if (!f && !g)
    {
      /* Since both are false, we need to follow only one argument.
       */
      if (a < 0)		/* 'a' empty clauses */
	add_ele_to_core (a);
      else if (b < 0)		/* 'b' empty clause */
	add_ele_to_core (b);
      else
	{
	  assert (a < b);	/* 'a' derived first */
	  add_ele_to_core (a);
	}
    }
  else if (!f)
    {
      add_ele_to_core (a);
    }
  else if (!g)
    {
      add_ele_to_core (b);
    }
  else if (f == 1 && g == 1)
    {
      /* nothing to be done here */
    }
  else if (f == 1)
    {
      add_ele_to_core (b);
    }
  else if (g == 1)
    {
      add_ele_to_core (a);
    }
  else if (f == h)		/* also covers 'f == g' (then f == h) */
    {
#ifndef NDEBUG
      trivial = 0;
      redundant = 1;
#endif
      add_ele_to_core (a);
    }
  else if (g == h)
    {
#ifndef NDEBUG
      trivial = 0;
      redundant = 1;
#endif
      add_ele_to_core (b);
    }
  else
    {
#ifndef NDEBUG
      trivial = 0;
      redundant = 0;
#endif
      add_ele_to_core (a);
      add_ele_to_core (b);

    }

  assert (redundant == is_redundant_and (and));
  assert (trivial == is_trivial_and (and));
}

/*------------------------------------------------------------------------*/

static void
core_eles (void)
{
  int i;
  Ele *ele;

  assert (root > 0);
  assert (root <= num_eles);

  add_ele_to_core (root);

  for (i = root; i >= 0; i--)
    {
      if (!marked (marked_eles, i))
	continue;
      ele = (eles + i);
      switch (ele->type)
	{
	case CLAUSE:
	  break;
	case EX:
	  single_child (exs + ele->idx);
	  break;
	case AND:
	  two_children (ands + ele->idx);
	  break;
	default:
	  assert (0);
	}
    }
}

/*------------------------------------------------------------------------*/

static void
core_and_lines (void)
{
  int f, g, h, f0, f1, g0, g1, h0, h1, x;
  int i;
  Line *line;

  for (i = num_lines; i > 0; i--)
    {
      if (!marked (marked_and_lines, i))
	continue;

      line = and_lines + i;
      if (is_redundant_line (line))
	continue;

      f = line->f;
      g = line->g;
      h = line->h;

      add_node_to_core (f);
      add_node_to_core (g);
      add_node_to_core (h);

      cofactors (f, g, x, f0, g0, f1, g1);
      cofactor (h, x, h0, h1);

      if (f0 && g0)
	{
	  if (f0 == 1)
	    {
	      assert (is_trivial_step (f0, g0));
	      add_node_to_core (g0);
	    }
	  else if (g0 == 1)
	    {
	      assert (is_trivial_step (f0, g0));
	      add_node_to_core (f0);
	    }
	  else if (f0 == g0)
	    {
	      assert (is_trivial_step (f0, g0));
	      add_node_to_core (f0);
	    }
	  else if (f0 == h0)
	    {
	      assert (!is_trivial_step (f0, g0));
	      assert (is_redundant_step (f0, g0, h0));
	      add_node_to_core (f0);
	    }
	  else if (g0 == h0)
	    {
	      assert (!is_trivial_step (f0, g0));
	      assert (is_redundant_step (f0, g0, h0));
	      add_node_to_core (g0);
	    }
	  else
	    {
	      assert (!is_trivial_step (f0, g0));
	      assert (!is_redundant_step (f0, g0, h0));
	      add_and_line_to_core (step2idx (f0, g0));
	    }
	}

      if (f1 && g1)
	{
	  if (f1 == 1)
	    {
	      assert (is_trivial_step (f1, g1));
	      add_node_to_core (g1);
	    }
	  else if (g1 == 1)
	    {
	      assert (is_trivial_step (f1, g1));
	      add_node_to_core (f1);
	    }
	  else if (f1 == g1)
	    {
	      assert (is_trivial_step (f1, g1));
	      add_node_to_core (f1);
	    }
	  else if (f1 == h1)
	    {
	      assert (!is_trivial_step (f1, g1));
	      assert (is_redundant_step (f1, g1, h1));
	      add_node_to_core (f1);
	    }
	  else if (g1 == h1)
	    {
	      assert (!is_trivial_step (f1, g1));
	      assert (is_redundant_step (f1, g1, h1));
	      add_node_to_core (g1);
	    }
	  else
	    {
	      assert (!is_trivial_step (f1, g1));
	      assert (!is_redundant_step (f1, g1, h1));
	      add_and_line_to_core (step2idx (f1, g1));
	    }
	}
    }

  assert (num_core_lines ==
	  count_marked (marked_and_lines, size_marked_and_lines));

  fprintf (output,
	   "c %d core lines (%.0f%% of all %d lines)\n",
	   num_core_lines, PERCENT (num_core_lines, num_lines), num_lines);
}

/*------------------------------------------------------------------------*/

static void
core_or_lines (void)
{
  int f, g, h, f0, f1, g0, g1, x;
  int i;
  Line *line;

  for (i = num_or_lines; i > 0; i--)
    {
      if (!marked (marked_or_lines, i))
	continue;

      line = or_lines + i;

      f = line->f;
      g = line->g;
      h = line->h;

      add_node_to_core (f);
      add_node_to_core (g);
      if (h != f && h != g)
	add_node_to_core (h);

      cofactors (f, g, x, f0, g0, f1, g1);
      //cofactor (h, x, h0, h1);

      if (f0 && g0)
	{
	  if (f0 == 1)
	    {
	      assert (is_trivial_step (f0, g0));
	      add_node_to_core (g0);
	    }
	  else if (g0 == 1)
	    {
	      assert (is_trivial_step (f0, g0));
	      add_node_to_core (f0);
	    }
	  else if (f0 == g0)
	    {
	      assert (is_trivial_step (f0, g0));
	      add_node_to_core (f0);
	    }
	  else
	    {
	      assert (!is_trivial_step (f0, g0));
	      add_or_line_to_core (or_step2idx (f0, g0));
	    }
	}

      if (f1 && g1)
	{
	  if (f1 == 1)
	    {
	      assert (is_trivial_step (f1, g1));
	      add_node_to_core (g1);
	    }
	  else if (g1 == 1)
	    {
	      assert (is_trivial_step (f1, g1));
	      add_node_to_core (f1);
	    }
	  else if (f1 == g1)
	    {
	      assert (is_trivial_step (f1, g1));
	      add_node_to_core (f1);
	    }
	  else
	    {
	      assert (!is_trivial_step (f1, g1));
	      add_or_line_to_core (or_step2idx (f1, g1));
	    }
	}
    }

  assert (num_core_or_lines ==
	  count_marked (marked_or_lines, size_marked_or_lines));

  fprintf (output,
	   "c %d core or lines (%.0f%% of all %d lines)\n",
	   num_core_or_lines,
	   PERCENT (num_core_or_lines, num_or_lines), num_or_lines);
}

/*------------------------------------------------------------------------*/

static void
add_variable_to_core (int idx)
{
  check_legal_var_idx (idx);
  if (marked (marked_variables, idx))
    return;

  mark (marked_variables, idx);
  num_core_variables++;
}

/*------------------------------------------------------------------------*/

static void
core_nodes (void)
{
  Node *node;
  int i;

  for (i = num_nodes - 1; i > 2; i--)
    {
      if (!marked (marked_nodes, i))
	continue;

      node = nodes + i;

      add_variable_to_core (node->var);

      add_node_to_core (node->child[0]);
      add_node_to_core (node->child[1]);
    }

  assert (num_core_nodes == count_marked (marked_nodes, size_marked_nodes));

  fprintf (output,
	   "c %d core nodes (%.0f%% of all %d nodes)\n",
	   num_core_nodes, PERCENT (num_core_nodes, num_nodes), num_nodes);

  assert (num_core_variables ==
	  count_marked (marked_variables, size_marked_variables));

  fprintf (output,
	   "c %d core variables (%.0f%% of all %d variables)\n",
	   num_core_variables,
	   PERCENT (num_core_variables, max_variable), max_variable);
}

/*------------------------------------------------------------------------*/

static void
core (void)
{
  double start_core_time = current_time ();

  size_marked_clauses = ((num_clauses + 1) + 31) / 32;
  size_marked_ands = (num_ands + 31) / 32;
  size_marked_exs = (num_exs + 31) / 32;
  size_marked_eles = ((num_eles + 1) + 31) / 32;
  size_marked_nodes = ((num_nodes + 1) + 31) / 32;
  size_marked_and_lines = ((num_lines + 1) + 31) / 32;
  size_marked_or_lines = ((num_or_lines + 1) + 31) / 32;
  size_marked_variables = ((max_variable + 1) + 31) / 32;

  NEWN (marked_clauses, size_marked_clauses);
  NEWN (marked_ands, size_marked_ands);
  NEWN (marked_exs, size_marked_exs);
  NEWN (marked_eles, size_marked_eles);
  NEWN (marked_and_lines, size_marked_and_lines);
  NEWN (marked_or_lines, size_marked_or_lines);
  NEWN (marked_nodes, size_marked_nodes);
  NEWN (marked_variables, size_marked_variables);

  if (find_root ())
    {
      core_eles ();
      core_and_lines ();
      core_or_lines ();
      core_nodes ();
    }

  core_time = current_time () - start_core_time;
  core_time = (core_time < 0) ? 0 : core_time;
  fprintf (output, "c extracted core in %.2f seconds\n", core_time);
}

/*------------------------------------------------------------------------*/

static Line *
and_step2line (int f, int g)
{
  return and_lines + step2idx (f, g);
}

/*------------------------------------------------------------------------*/

static Line *
or_step2line (int f, int g)
{
  return or_lines + or_step2idx (f, g);
}

/*------------------------------------------------------------------------*/

static int
and_line2chain (Line * line)
{
  int idx, res;

  assert (first_and_line_chain);

  idx = line - and_lines;
  check_legal_line_idx (idx);
  res = 2 * idx + first_and_line_chain - 1;

  return res;
}

/*------------------------------------------------------------------------*/

/* line f->h */
static int
or_line2chain1 (Line * line)
{
  int idx, res;

  assert (first_or_line_chain);

  idx = line - or_lines;
  check_legal_or_line_idx (idx);
  res = 4 * idx + first_or_line_chain - 3;

  return res;
}

/* line g->h */
static int
or_line2chain2 (Line * line)
{
  return or_line2chain1 (line) + 2;
}

/*------------------------------------------------------------------------*/
#ifdef EBDDRES_DUMP
/*------------------------------------------------------------------------*/

static void
dump_line (Line * l)
{
  int idx = l - and_lines;

  fprintf (output, "c line %d: %d & %d = %d", idx, l->f, l->g, l->h);

  if (first_and_line_chain)
    fprintf (output,
	     " chain %d %d", and_line2chain (l), and_line2chain (l) - 1);

  if (is_redundant_line (l))
    fputs (" redundant", output);

  if (marked_and_lines && !marked (marked_and_lines, idx))
    fputs (" not-in-core", output);

  fputc ('\n', output);
}

/*------------------------------------------------------------------------*/

static void
dump_or_line (Line * l)
{
  int idx = l - or_lines;

  fprintf (output, "c line %d: %d & %d = %d", idx, l->f, l->g, l->h);

  if (first_or_line_chain)
    fprintf (output, " chain %d %d", or_line2chain2 (l),
	     or_line2chain2 (l) - 1);

  if (is_redundant_line (l))
    fputs (" redundant", output);

  if (marked_or_lines && !marked (marked_or_lines, idx))
    fputs (" not-in-core", output);

  fputc ('\n', output);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static int
step2chain (int f, int g)
{
  return and_line2chain (and_step2line (f, g));
}

/*------------------------------------------------------------------------*/

static int
or_step2chain1 (int f, int g)
{
  return or_line2chain1 (or_step2line (f, g));
}

/*------------------------------------------------------------------------*/

static int
or_step2chain2 (int f, int g)
{
  return or_line2chain2 (or_step2line (f, g));
}

/*------------------------------------------------------------------------*/

static int
ex2chain (int idx)
{
  return first_ex_chain + 2 * (idx + 1) - 1;
}

/*------------------------------------------------------------------------*/

static void
trace_and_line (Line * line)
{
  int fdef, gdef, hdef;
  int f1, f0, g1, g0, h1, h0, chain0, chain1;
  int f = line->f, g = line->g, h = line->h;
  int antecedents[4], i, count;
  int x;

  assert (h != 1);
  assert (f < g);
  assert (!is_trivial_step (f, g));

  antecedents[0] = 0;

  chain0 = ++traced_chains;
  chain1 = ++traced_chains;

  if (!no_core && !marked (marked_and_lines, line - and_lines))
    return;

  assert (step2chain (f, g) == chain1);

  if (h == g)
    return;

  if (h == f)
    return;

  assert (!is_redundant_line (line));

  cofactors (f, g, x, f0, g0, f1, g1);
  cofactor (h, x, h0, h1);

  assert (h0 != g0 || h1 != g1);
  assert (h0 != f0 || h1 != f1);

  fdef = node2def (f);
  gdef = node2def (g);
  hdef = node2def (h);

  count = 0;

  if (f0 == h0)
    {
#ifndef NDEBUG
      if (!is_trivial_step (f0, g0))
	assert (is_redundant_line (and_step2line (f0, g0)));
#endif
      if (f != f0)
	antecedents[count++] = node2chain (f, -1, 1);
    }
  else if (g0 == h0)
    {
#ifndef NDEBUG
      if (!is_trivial_step (f0, g0))
	assert (is_redundant_line (and_step2line (f0, g0)));
#endif
      if (g != g0)
	antecedents[count++] = node2chain (g, -1, 1);
    }
  else
    {
      assert (f0 != h0 && g0 != h0);
      assert (!is_trivial_step (f0, g0));
      assert (!is_redundant_line (and_step2line (f0, g0)));

      antecedents[count++] = step2chain (f0, g0);

      if (f != f0)
	antecedents[count++] = node2chain (f, -1, 1);

      if (g != g0)
	antecedents[count++] = node2chain (g, -1, 1);
    }

  if (h != 0 && h != h0)
    antecedents[count++] = node2chain (h, 1, 1);

  assert (count >= 1);

  if (count > 1)
    {
      putispace (chain0, trace_file);

      if (extended_trace_format)
	{
	  if (f0 == h0)
	    put4ispace (-fdef, hdef, x, 0, trace_file);
	  else if (g0 == h0)
	    put4ispace (-gdef, hdef, x, 0, trace_file);
	  else
	    put5ispace (-fdef, -gdef, hdef, x, 0, trace_file);
	}
      else
	fputs ("* ", trace_file);

      for (i = 0; i < count; i++)
	putispace (antecedents[i], trace_file);

      fputs ("0\n", trace_file);
    }
  else
    chain0 = antecedents[0];	/* no 'chain0' necessary */

  putispace (chain1, trace_file);

  if (extended_trace_format)
    put4ispace (-fdef, -gdef, hdef, 0, trace_file);
  else
    fputs ("* ", trace_file);

  if (f1 == h1)
    {
#ifndef NDEBUG
      if (!is_trivial_step (f1, g1))
	assert (is_redundant_line (and_step2line (f1, g1)));
#endif
      if (f != f1)
	putispace (node2chain (f, -1, -1), trace_file);
    }
  else if (g1 == h1)
    {
#ifndef NDEBUG
      if (!is_trivial_step (f1, g1))
	assert (is_redundant_line (and_step2line (f1, g1)));
#endif
      if (g != g1)
	putispace (node2chain (g, -1, -1), trace_file);
    }
  else
    {
      assert (f1 != h1 && g1 != h1);
      assert (!is_trivial_step (f1, g1));
      assert (!is_redundant_line (and_step2line (f1, g1)));

      putispace (step2chain (f1, g1), trace_file);

      if (f != f1)
	putispace (node2chain (f, -1, -1), trace_file);

      if (g != g1)
	putispace (node2chain (g, -1, -1), trace_file);
    }

  if (h != 0 && h != h1)
    putispace (node2chain (h, 1, -1), trace_file);

  putispace (chain0, trace_file);
  fputs ("0\n", trace_file);

  if (extended_trace_format)
    fputc ('\n', trace_file);
}

/*------------------------------------------------------------------------*/

static int
find_any_line2 (int f, const int g, const int h)
{
  int p, idx;
  int t1 = f, t2 = g;
  Line *line;
  Node n;

  if (nodes[g].var == -1 || g == h || is_trivial_step (f, g))
    return 0;

  normalize (t1, t2);
  p = find_or_line (t1, t2);
  idx = or_cache[p];
  line = or_lines + idx;
  if (line->h == h)
    return idx;

  n = nodes[f];
  n = nodes[n.child[0]];
  if (n.var != -1)
    {
      idx = find_any_line2 (nodes[f].child[0], g, h);
      if (idx)
	return idx;
    }

  n = nodes[f];
  n = nodes[n.child[1]];
  if (n.var != -1)
    return find_any_line2 (nodes[f].child[1], g, h);

  return 0;
}

/*------------------------------------------------------------------------*/

static int
find_any_line (const int f, int g, const int h)
{
  int p, idx;
  int t1 = f, t2 = g;
  Line *line;
  Node n;

  if (nodes[f].var == -1 || f == h || is_trivial_step (f, g))
    return 0;

  normalize (t1, t2);
  p = find_or_line (t1, t2);
  idx = or_cache[p];
  line = or_lines + idx;
  if (line->h == h)
    return idx;

  n = nodes[g];
  n = nodes[n.child[0]];
  if (n.var != -1)
    {
      idx = find_any_line (f, nodes[g].child[0], h);
      if (idx)
	return idx;
    }

  n = nodes[g];
  n = nodes[n.child[1]];
  if (n.var != -1)
    return find_any_line (f, nodes[g].child[1], h);

  return 0;
}

/*------------------------------------------------------------------------*/

static void
trace_or_line (Line * line)
{
  int fdef, gdef, hdef;
  int f1, f0, g1, g0, h1, h0, chain0, chain1, chain2, chain3, other_branch;
  int f = line->f, g = line->g, h = line->h;
  int x;
  Line *l;

  assert (h);
  assert (f < g);
  assert (!is_trivial_step (f, g));

  chain0 = ++traced_chains;
  chain1 = ++traced_chains;
  chain2 = ++traced_chains;
  chain3 = ++traced_chains;

  if (!no_core && !marked (marked_or_lines, line - or_lines))
    return;

  if (h == 1)
    return;

  /* 1st part, prove that f -> h */

  cofactors (f, h, x, f0, h0, f1, h1);
  cofactor (g, x, g0, g1);

  fdef = node2def (f);
  hdef = node2def (h);

  if (f == h)
    goto GH;

  if (cmp_var (x, nodes[g].var) < 0)
    {
      l = or_lines + find_any_line (f0, g0, h0);
      g0 = (f0 == l->f) ? l->g : l->f;

      l = or_lines + find_any_line (f1, g1, h1);
      g1 = (f1 == l->f) ? l->g : l->f;
    }

  /* branch F0 */

  other_branch = 0;
  if (h0 != 1 && f0 != 0 && f0 != 1)
    {
      putispace (chain0, trace_file);

      if (extended_trace_format)
	put4ispace (-fdef, hdef, x, 0, trace_file);
      else
	fputs ("* ", trace_file);

      if (f0 != h0 && !is_trivial_step (f0, g0))
	{
	  if (f0 < g0)
	    putispace (or_step2chain1 (f0, g0), trace_file);
	  else
	    putispace (or_step2chain2 (f0, g0), trace_file);
	}

      if (f != f0)
	putispace (node2chain (f, -1, 1), trace_file);
      if (h != h0)
	putispace (node2chain (h, 1, 1), trace_file);

      fputs ("0\n", trace_file);
    }
  else if (h0 == 1)
    other_branch = node2chain (h, 1, 1);
  else if (f0 == 0)
    other_branch = node2chain (f, -1, 1);
  else if (f0 == 1)
    other_branch = node2chain (h, 1, 1);

  /* branch F1 */

  putispace (chain1, trace_file);
  if (extended_trace_format)
    put3ispace (-fdef, hdef, 0, trace_file);
  else
    fputs ("* ", trace_file);

  if (other_branch)
    put2ispace (true_chain, other_branch, trace_file);
  else
    putispace (chain0, trace_file);

  if (h1 != 1 && f1 != 0 && f1 != 1)
    {
      if (f1 != h1 && !is_trivial_step (f1, g1))
	{
	  if (f1 < g1)
	    putispace (or_step2chain1 (f1, g1), trace_file);
	  else
	    putispace (or_step2chain2 (f1, g1), trace_file);
	}

      if (f != f1)
	putispace (node2chain (f, -1, -1), trace_file);
      if (h != h1)
	putispace (node2chain (h, 1, -1), trace_file);
    }
  else if (h1 == 1)
    {
      if (!other_branch)
	putispace (true_chain, trace_file);
      putispace (node2chain (h, 1, -1), trace_file);
    }
  else if (f1 == 0)
    {
      if (!other_branch)
	putispace (true_chain, trace_file);
      putispace (node2chain (f, -1, -1), trace_file);
    }
  else if (f1 == 1)
    putispace (node2chain (h, 1, -1), trace_file);

  fputs ("0\n", trace_file);

  /* 2nd part, prove that g->h */

GH:
  cofactors (g, h, x, g0, h0, g1, h1);
  cofactor (f, x, f0, f1);

  gdef = node2def (g);
  hdef = node2def (h);

  if (g == h)
    return;

  if (cmp_var (x, nodes[f].var) < 0)
    {
      l = or_lines + find_any_line2 (f0, g0, h0);
      f0 = (g0 == l->g) ? l->f : l->g;

      l = or_lines + find_any_line2 (f1, g1, h1);
      f1 = (g1 == l->g) ? l->f : l->g;
    }

  /* branch G0 */

  other_branch = 0;
  if (h0 != 1 && g0 != 0 && g0 != 1)
    {
      putispace (chain2, trace_file);

      if (extended_trace_format)
	put4ispace (-gdef, hdef, x, 0, trace_file);
      else
	fputs ("* ", trace_file);

      if (g0 != h0 && !is_trivial_step (f0, g0))
	{
	  if (f0 < g0)
	    putispace (or_step2chain2 (f0, g0), trace_file);
	  else
	    putispace (or_step2chain1 (f0, g0), trace_file);
	}

      if (g != g0)
	putispace (node2chain (g, -1, 1), trace_file);
      if (h != h0)
	putispace (node2chain (h, 1, 1), trace_file);

      fputs ("0\n", trace_file);
    }
  else if (h0 == 1)
    other_branch = node2chain (h, 1, 1);
  else if (g0 == 0)
    other_branch = node2chain (g, -1, 1);
  else if (g0 == 1)
    other_branch = node2chain (h, 1, 1);

  /* branch G1 */

  putispace (chain3, trace_file);
  if (extended_trace_format)
    put3ispace (-gdef, hdef, 0, trace_file);
  else
    fputs ("* ", trace_file);

  if (other_branch)
    put2ispace (true_chain, other_branch, trace_file);
  else
    putispace (chain2, trace_file);

  if (h1 != 1 && g1 != 0 && g1 != 1)
    {
      if (g1 != h1 && !is_trivial_step (f1, g1))
	{
	  if (f1 < g1)
	    putispace (or_step2chain2 (f1, g1), trace_file);
	  else
	    putispace (or_step2chain1 (f1, g1), trace_file);
	}

      if (g != g1)
	putispace (node2chain (g, -1, -1), trace_file);
      if (h != h1)
	putispace (node2chain (h, 1, -1), trace_file);
    }
  else if (h1 == 1)
    {
      if (!other_branch)
	putispace (true_chain, trace_file);
      putispace (node2chain (h, 1, -1), trace_file);
    }
  else if (g1 == 0)
    {
      if (!other_branch)
	putispace (true_chain, trace_file);
      putispace (node2chain (g, -1, -1), trace_file);
    }
  else if (g1 == 1)
    putispace (node2chain (h, 1, -1), trace_file);
  fputs ("0\n", trace_file);


}

/*------------------------------------------------------------------------*/

static void
trace_ex (Ex * ex)
{
  int chain0, chain1;
  int fdef, exfdef;
  Node *n;

  if (!marked (marked_exs, ex - exs))
    return;

  fdef = node2def (ele2bdd (ex->a));
  exfdef = node2def (ex->h);

  chain0 = ++traced_chains;

  if (is_trivial_ex (ex))
    {
      chain1 = ++traced_chains;
      putispace (chain1, trace_file);

      if (extended_trace_format)
	put3ispace (-fdef, exfdef, 0, trace_file);
      else
	fputs ("* ", trace_file);

      put2ispace (true_chain, node2chain (ele2bdd (ex->a), -1, -1),
		  trace_file);
      putispace (node2chain (ele2bdd (ex->a), -1, 1), trace_file);
      fputs ("0\n", trace_file);
      return;
    }

  putispace (chain0, trace_file);

  if (extended_trace_format)
    put4ispace (-fdef, exfdef, (nodes + ele2bdd (ex->a))->var, 0, trace_file);
  else
    fputs ("* ", trace_file);

  n = nodes + ele2bdd (ex->a);

  putispace (node2chain (ele2bdd (ex->a), -1, 1), trace_file);
  /* line f->h */
  if (n->child[0] != ex->h)
    {
      if (n->child[0] < n->child[1])
	putispace (or_line2chain1 (or_step2line (n->child[0], n->child[1])),
		   trace_file);
      else
	putispace (or_line2chain2 (or_step2line (n->child[0], n->child[1])),
		   trace_file);
    }

  fputs ("0\n", trace_file);

  chain1 = ++traced_chains;
  putispace (chain1, trace_file);

  if (extended_trace_format)
    put3ispace (-fdef, exfdef, 0, trace_file);
  else
    fputs ("* ", trace_file);

  putispace (chain0, trace_file);

  putispace (node2chain (ele2bdd (ex->a), -1, -1), trace_file);
  /* line g->h */
  if (n->child[1] != ex->h)
    {
      if (n->child[0] < n->child[1])
	putispace (or_line2chain2 (or_step2line (n->child[0], n->child[1])),
		   trace_file);
      else
	putispace (or_line2chain1 (or_step2line (n->child[0], n->child[1])),
		   trace_file);
    }

  fputs ("0\n", trace_file);


  return;
}

/*------------------------------------------------------------------------*/

static void
trace_core_clauses (void)
{
  int i;
  for (i = 1; i <= num_clauses; i++)
    if (marked (marked_clauses, i))
      putispace (clause2unit (i), trace_file);
}

/*------------------------------------------------------------------------*/

static void
trace_core_ands (void)
{
  int i, f, g;
  And *and;

  for (i = 0; i < num_ands; i++)
    {
      if (!marked (marked_ands, i))
	continue;

      and = ands + i;

      if (is_trivial_and (and))
	continue;

      if (is_redundant_and (and))
	continue;

      f = ele2bdd (and->a);
      g = ele2bdd (and->b);

      putispace (step2chain (f, g), trace_file);
    }
}

/*------------------------------------------------------------------------*/

static void
trace_core_exs (void)
{
  int i;

  for (i = 0; i < num_exs; i++)
    {
      if (!marked (marked_exs, i))
	continue;

      putispace (ex2chain (i), trace_file);
    }
}

/*------------------------------------------------------------------------*/

static int
ex_unit (int idx)
{
  return first_ex_unit_chain + (2 * idx) + 1;
}

/*------------------------------------------------------------------------*/

static void
line_items (int idx)
{
  Ele *e = eles + idx;
  And *and;
  int f, g;

  switch (e->type)
    {
    case CLAUSE:
      if (!marked (marked_clauses, e->idx))
	break;
      putispace (clause2unit (e->idx), trace_file);
      break;
    case AND:
      if (!marked (marked_ands, e->idx))
	break;
      and = ands + e->idx;
      f = ele2bdd (and->a);
      g = ele2bdd (and->b);
      if (!is_trivial_and (and) && !is_redundant_and (and))
	putispace (step2chain (f, g), trace_file);
      if (f != 0 && f != 1)
	line_items (and->a);
      if (g != 0 && g != 1)
	line_items (and->b);
      break;
    case EX:
      putispace (ex_unit (e->idx), trace_file);
      break;
    }
  return;
}

/*------------------------------------------------------------------------*/

static void
trace_ex_unit (Ex * ex)
{
  int chain1, chain2;

  chain1 = ++traced_chains;
  chain2 = ++traced_chains;

  if (!marked (marked_exs, ex - exs))
    return;

  putispace (chain1, trace_file);
  fputs ("* ", trace_file);

  line_items (ex->a);
  fputs ("0\n", trace_file);

  putispace (chain2, trace_file);
  fputs ("* ", trace_file);
  put2ispace (chain1, ex2chain (ex - exs), trace_file);
  fputs ("0\n", trace_file);
}

/*------------------------------------------------------------------------*/

static void
trace_root (void)
{
  int idx;
  And *and;

  traced_chains++;

  assert (root_chain == traced_chains);
  putispace (root_chain, trace_file);

  idx = ele2bdd (root);
  if (idx)
    put2ispace (node2def (idx), 0, trace_file);
  else
    put2ispace (0, true_chain, trace_file);

  if (strategy == 0 || strategy == 1)
    {

      trace_core_clauses ();
      trace_core_ands ();
      trace_core_exs ();
    }
  else
    {
      and = ands + (eles + root)->idx;
      if (!is_trivial_and (and) && !is_redundant_and (and))
	line_items (root);

    }
  fputs ("0\n", trace_file);
}

/*------------------------------------------------------------------------*/

static void
trace_empty_clause (void)
{
  first_clause_chain = 1;
  assert (empty_clause);
  print_original_clause_as_chain (empty_clause, trace_file);
  traced_chains = first_clause_chain + num_clauses;

  put4i (traced_chains, 0, clause2chain (empty_clause), 0, trace_file);
}

/*------------------------------------------------------------------------*/

static void
trace (void)
{
  double trace_start_time = current_time ();

  assert (!empty_clause);

  assert (!traced_chains);
  first_clause_chain = traced_chains + 1;
  forall_clauses (trace_clause);
  assert (traced_chains == first_clause_chain + num_clauses - 1);

  true_variable = max_variable + 1;
  traced_variables = true_variable;
  true_chain = ++traced_chains;
  if (extended_trace_format)
    fputc ('\n', trace_file);
  put4i (true_chain, true_variable, 0, 0, trace_file);
  fputc ('\n', trace_file);
  if (extended_trace_format)
    fputc ('\n', trace_file);

  first_node_variable = true_variable + 1;
  first_node_chain = traced_chains + 1;
  forall_bdds (trace_node);

  assert (traced_chains == first_node_chain + 4 * (num_nodes - 2) - 1);
  assert (traced_variables == first_node_variable + num_nodes - 3);

  first_root_unit_chain = traced_chains + 1;
  forall_clauses (trace_root_unit_clause);
  assert (traced_chains == first_root_unit_chain + num_clauses - 1);

  fputc ('\n', trace_file);

  first_and_line_chain = traced_chains + 1;
  forall_lines (trace_and_line);
  assert (traced_chains == first_and_line_chain + 2 * num_lines - 1);

  fputc ('\n', trace_file);

  first_or_line_chain = traced_chains + 1;
  forall_or_lines (trace_or_line);

  fputc ('\n', trace_file);

  first_ex_chain = traced_chains + 1;
  forall_exs (trace_ex);

  fputc ('\n', trace_file);

  first_ex_unit_chain = traced_chains + 1;
  forall_exs (trace_ex_unit);

  fputc ('\n', trace_file);
  if (root != TRUE_AND)
    {
      root_chain = traced_chains + 1;
      trace_root ();
      assert (root_chain == traced_chains);
    }

  fflush (trace_file);
  trace_time = current_time () - trace_start_time;
  trace_time = (trace_time < 0) ? 0 : trace_time;
  fprintf (output,
	   "c\nc generated %s proof trace in %.2f seconds\n",
	   extended_trace_format ? "extended" : "compact", trace_time);
}

/*------------------------------------------------------------------------*/

static void
dump ()
{
#ifdef EBDDRES_DUMP
  if (!dump_all)
    return;

  dump_variable_order ();
  dump_variable_elim ();
  dump_buckets ("dump: ");
  dump_occurrences ();
  forall_variables (dump_variable);

  forall_clauses (dump_clause);
  forall_bdds (dump_node);
  forall_lines (dump_line);
  forall_or_lines (dump_or_line);
  forall_ands (dump_and);
  forall_eles (dump_ele);
#endif
}

/*------------------------------------------------------------------------*/

static void
stats (void)
{
  int num_trivial_steps, num_redundant_steps;

  fputs ("c\n", output);

  fprintf (output,
	   "c %d nodes (%.2f average number collisions)\n",
	   num_nodes - 2, table_finds ? table_collisions / table_finds : 0.0);

  fprintf (output,
	   "c %d lines (%d redundant, %.2f average number collisions)\n",
	   num_lines, num_redundant_lines,
	   cache_finds ? cache_collisions / cache_finds : 0.0);

  fprintf (output,
	   "c %d orlines (%d redundant, %.2f average number collisions)\n",
	   num_or_lines, num_redundant_or_lines,
	   cache_finds ? cache_collisions / cache_finds : 0.0);

  fprintf (output,
	   "c %d ands (%d trivial, %d redundant)\n",
	   num_ands, num_trivial_ands, num_redundant_ands);

  fprintf (output,
	   "c %d exs (%d trivial, %d redundant)\n",
	   num_exs, num_trivial_exs, num_redundant_exs);

  num_trivial_steps = num_steps - num_lines - num_or_lines;
  num_redundant_steps =
    num_trivial_steps + num_redundant_lines + num_redundant_or_lines;

  assert (num_steps >= num_redundant_steps);
  assert (num_redundant_steps >= num_trivial_steps);
  assert (num_trivial_steps >= 0);

  fprintf (output,
	   "c %.0f steps (%d trivial, %d redundant)\n",
	   num_steps, num_trivial_steps, num_redundant_steps);

  fprintf (output,
	   "c %.0f%% cache hit ratio\n",
	   100.0 * (cache_lookups ? cache_hits / cache_lookups : 0.0));

  fputs ("c\n", output);

  fprintf (output,
	   "c %.2f seconds (%.2f sat + %.2f core + %.2f trace)\n",
	   seconds (), sat_time, core_time, trace_time);

  fprintf (output, "c %.1f MB\n", mb ());
}

/*------------------------------------------------------------------------*/

#define INIT (1 << 10)

/*------------------------------------------------------------------------*/

#define USAGE \
"usage: ebddres [<option> ...]\n" \
"\n" \
"with <option> one of the following\n" \
"\n" \
"  -h                print this command line option summary\n" \
"  --dump            dump debugging information about lines and nodes\n" \
"  --config          print compile time configuration\n" \
"  -t <trace>        dump compact proof trace\n" \
"  -T <trace>        dump extended proof trace\n" \
"  --original-order  use original variable order (default)\n" \
"  --reverse-order   use reverse variable order\n" \
"  --dfs-order       use DFS variable order\n" \
"  --sum-order       use number of occurrences for ordering\n" \
"  --no-force        dot not optimize variable order with FORCE\n" \
"  --no-core         do not extract core and trace everything\n" \
"  --schedule        try to optimize schedule of clauses\n" \
"  --tree            tree shaped instead of linear computation\n" \
"  --ve              variable elimination\n" \
"  --vetree          variable elimination, tree computation\n" \
"  --vetrial         elim multiple variables\n" \
"  -o <output>       set output file\n" \
"  <input>           input file in DIMACS format\n" \
"  -w <order-output-file>\n"

/*------------------------------------------------------------------------*/

int
main (int argc, char **argv)
{
  const char *msg, *trace_file_name;
  int i, err, done, res;
  FILE *file;

  start_time = current_time ();
  trace_file_name = 0;

  output = stdout;
  close_output = 0;

  input = stdin;
  input_name = "<stdin>";
  close_input = 0;
  lineno = 1;

  NEWN (literals, size_literals = INIT);
  NEWN (clauses, size_clauses = INIT);
  NEWN (table, size_table = INIT);
  NEWN (nodes, size_nodes = INIT);
  NEWN (and_cache, size_and_cache = INIT);
  NEWN (or_cache, size_or_cache = INIT);
  NEWN (and_lines, size_and_lines = INIT);
  NEWN (or_lines, size_or_lines = INIT);
  NEWN (eles, size_eles = INIT);
  NEWN (ands, size_ands = INIT);
  NEWN (stack, size_stack = INIT);

  nodes[0].var = -1;
  nodes[0].child[0] = nodes[0].child[1] = 0;

  nodes[1].var = -1;
  nodes[1].child[0] = nodes[0].child[1] = 1;

  num_nodes = 2;

  res = err = done = 0;

  for (i = 1; !err && !done && i < argc; i++)
    {
      if (!strcmp (argv[i], "-h"))
	{
	  fputs (USAGE, output);
	  done = 1;
	}
      else if (!strcmp (argv[i], "--version"))
	{
	  fprintf (output, "%s\n", EBDDRES_VERSION);
	  done = 1;
	}
      else if (!strcmp (argv[i], "--schedule"))
	orig_schedule = 0;
      else if (!strcmp (argv[i], "--tree"))
	strategy = 1;
      else if (!strcmp (argv[i], "--ve"))
	strategy = 2;
      else if (!strcmp (argv[i], "--vetree"))
	strategy = 3;
      else if (!strcmp (argv[i], "--vetrial"))
	{
	  strategy = 4;
	  elim_tresh = atoi (argv[++i]);
	}
      else if (!strcmp (argv[i], "--original-order"))
	init_order_function = init_original_order;
      else if (!strcmp (argv[i], "--reverse-order"))
	init_order_function = init_reverse_order;
      else if (!strcmp (argv[i], "--dfs-order"))
	init_order_function = init_dfs_order;
      else if (!strcmp (argv[i], "--sum-order"))
	init_order_function = init_sum_order;
      else if (!strcmp (argv[i], "--no-force"))
	optimize_order = 0;
      else if (!strcmp (argv[i], "--no-core"))
	no_core = 1;
      else if (!strcmp (argv[i], "--config"))
	{
	  /**INDENT-OFF**/
	  fprintf  (output,
	  "ebddres version " EBDDRES_VERSION "\n"
	  "$Id$\n"
	  EBDDRES_OS "\n"
	  "CC=" EBDDRES_CC "\n"
	  "CFLAGS=" EBDDRES_CFLAGS "\n"
#ifdef EBDDRES_DEBUG
	  "debug=yes\n"
#else
	  "debug=no\n"
#endif
#ifdef EBDDRES_DUMP
	  "dump=yes\n"
#else
	  "dump=no\n"
#endif
	  );
	  /**INDENT-ON**/
	  done = 1;
	}
      else if (!strcmp (argv[i], "--id"))
	{
	  fprintf (output,
		   "$Id$\n");
	  done = 1;
	}
      else if (!strcmp (argv[i], "--dump"))
	{
#ifdef EBDDRES_DUMP
	  dump_all = 1;
#endif
	}
      else if (!strcmp (argv[i], "-o"))
	{
	  if (close_output)
	    err = die ("multiple output files");
	  else if (i + 1 == argc)
	    err = die ("argument to '-o' missing");
	  else if (!(file = fopen (argv[++i], "w")))
	    err = die ("can not write to '%s'", argv[i]);
	  else
	    {
	      output = file;
	      close_output = 1;
	    }
	}
      else if (!strcmp (argv[i], "-t") || !strcmp (argv[i], "-T"))
	{
	  if (trace_file_name)
	    err = die ("multiple trace files");
	  else if (i + 1 == argc)
	    err = die ("argument to '-%c' missing", argv[i][1]);
	  else
	    trace_file_name = argv[++i];

	  extended_trace_format = (argv[i - 1][1] == 'T');
	}
      else if (!strcmp (argv[i], "-w"))
	{
	  if (order_file)
	    err = die ("multiple order output files");
	  else if (i + 1 == argc)
	    err = die ("argument to '-w' missing");
	  else if (!(file = fopen (argv[++i], "w")))
	    err = die ("can not write to '%s'", argv[i]);
	  else
	    order_file = file;
	}
      else if (argv[i][0] == '-')
	err = die ("unknown command line option '%s'", argv[i]);
      else if (close_input)
	err = die ("multiple input files");
      else if (!(file = fopen (argv[i], "r")))
	err = die ("can not read '%s'", argv[i]);
      else
	{
	  input_name = argv[i];
	  close_input = 1;
	  input = file;
	}
    }

  if (!err && !done)
    {
      if ((msg = parse ()))
	{
	  fprintf (output, "%s:%d: %s\n", input_name, lineno, msg);
	  err = 1;
	}
      else
	{
	  res = sat ();

	  fputs ("c\n", output);
	  if (res == SATISFIABLE)
	    {
	      fputs ("s SATISFIABLE\n", output);

	      if (strategy == 0 || strategy == 1)
		extract_model ();
	      else
		extract_model_ve ();

	      print_assignment ();
	    }
	  else if (res == UNSATISFIABLE)
	    {
	      fputs ("s UNSATISFIABLE\n", output);
	    }
	  else
	    fputs ("s UNKNOWN\n", output);
	}

      if (trace_file_name)
	{
	  if ((trace_file = fopen (trace_file_name, "w")))
	    {
	      if (empty_clause)
		trace_empty_clause ();
	      else
		{
		  core ();
		  trace ();
		}

	      fclose (trace_file);
	    }
	  else
	    err = die ("can not write to '%s'", trace_file_name);
	}

      if (!err)
	dump ();
    }

  if (order_file)
    fclose (order_file);

  if (close_input)
    fclose (input);

  DELETEN (literals, size_literals);
  DELETEN (clauses, size_clauses);
  DELETEN (table, size_table);
  DELETEN (nodes, size_nodes);
  DELETEN (and_cache, size_and_cache);
  DELETEN (or_cache, size_or_cache);
  DELETEN (and_lines, size_and_lines);
  DELETEN (or_lines, size_or_lines);
  DELETEN (eles, size_eles);
  DELETEN (ands, size_ands);
  DELETEN (exs, size_exs);
  DELETEN (stack, size_stack);

  if (strategy == 2 || strategy == 3)
    DELETEN (buckets, size_buckets);

  if (marked_nodes)
    DELETEN (marked_nodes, size_marked_nodes);

  if (marked_and_lines)
    DELETEN (marked_and_lines, size_marked_and_lines);

  if (marked_or_lines)
    DELETEN (marked_or_lines, size_marked_or_lines);

  if (marked_variables)
    DELETEN (marked_variables, size_marked_variables);

  if (schedule)
    DELETEN (schedule, num_clauses);

  if (clause_pos)
    DELETEN (clause_pos, num_clauses + 1l);

  if (variable_widths)
    DELETEN (variable_widths, max_variable + 1l);

  if (variables)
    {
      DELETEN (variables, max_variable + 1l);
      DELETEN (variable_order, max_variable + 1l);
      DELETEN (variable_elim, max_variable + 1l);
    }

  if (occurrences)
    DELETEN (occurrences, size_occurrences);

  if (marked_ands)
    DELETEN (marked_ands, size_marked_ands);

  if (marked_eles)
    DELETEN (marked_eles, size_marked_eles);

  if (marked_exs)
    DELETEN (marked_exs, size_marked_exs);

  if (marked_clauses)
    DELETEN (marked_clauses, size_marked_clauses);

  if (!done && !err)
    stats ();

  if (close_output)
    fclose (output);

  assert (getenv ("LEAK") || !current_bytes);

  return res;
}
